import base64
import csv
import hashlib
import io
import json
import jsonlines as jsonl
import os
import re
import string
import unicodedata
import yaml
from datetime import datetime
from io import BytesIO
from PIL import Image
from typing import Tuple

from evalscope.constants import DumpMode
from evalscope.utils.logger import get_logger

logger = get_logger()


class OutputsStructure:
    LOGS_DIR = 'logs'
    PREDICTIONS_DIR = 'predictions'
    REVIEWS_DIR = 'reviews'
    REPORTS_DIR = 'reports'
    CONFIGS_DIR = 'configs'

    def __init__(self, outputs_dir: str, is_make=True):
        self.outputs_dir = outputs_dir
        self.is_make = is_make
        self._dirs = {
            'logs_dir': None,
            'predictions_dir': None,
            'reviews_dir': None,
            'reports_dir': None,
            'configs_dir': None
        }

    def _get_dir(self, attr_name, dir_name) -> str:
        if self._dirs[attr_name] is None:
            dir_path = os.path.join(self.outputs_dir, dir_name)
            if self.is_make:
                os.makedirs(dir_path, exist_ok=True)
            self._dirs[attr_name] = dir_path
        return self._dirs[attr_name]

    @property
    def logs_dir(self):
        return self._get_dir('logs_dir', OutputsStructure.LOGS_DIR)

    @property
    def predictions_dir(self):
        return self._get_dir('predictions_dir', OutputsStructure.PREDICTIONS_DIR)

    @property
    def reviews_dir(self):
        return self._get_dir('reviews_dir', OutputsStructure.REVIEWS_DIR)

    @property
    def reports_dir(self):
        return self._get_dir('reports_dir', OutputsStructure.REPORTS_DIR)

    @property
    def configs_dir(self):
        return self._get_dir('configs_dir', OutputsStructure.CONFIGS_DIR)


def jsonl_to_list(jsonl_file):
    """
    Read jsonl file to list.

    Args:
        jsonl_file: jsonl file path.

    Returns:
        list: list of lines. Each line is a dict.
    """
    try:
        res_list = []
        with jsonl.open(jsonl_file, mode='r') as reader:
            for line in reader.iter(type=dict, allow_none=True, skip_invalid=False):
                res_list.append(line)
    except Exception:
        # Fallback to reading line by line
        res_list = []
        with open(jsonl_file, 'r', encoding='utf-8') as f:
            for line in f:
                if line.strip():  # Skip empty lines
                    res_list.append(json.loads(line.strip()))
    if not res_list:
        logger.warning(f'No data found in {jsonl_file}.')
    return res_list


def jsonl_to_reader(jsonl_file):
    """
    Read jsonl file to reader object.

    Args:
        jsonl_file: jsonl file path.

    Returns:
        reader: jsonl reader object.
    """
    with jsonl.open(jsonl_file, mode='r') as reader:
        return reader


def dump_jsonl_data(data_list, jsonl_file, dump_mode=DumpMode.OVERWRITE):
    """
    Dump data to jsonl file.

    Args:
        data_list: data list to be dumped.  [{'a': 'aaa'}, ...]
        jsonl_file: jsonl file path.
        dump_mode: dump mode. It can be 'overwrite' or 'append'.
    """
    if not jsonl_file:
        raise ValueError('output file must be provided.')

    jsonl_file = os.path.expanduser(jsonl_file)

    if not isinstance(data_list, list):
        data_list = [data_list]

    # Convert non-serializable types to serializable ones
    data_list = convert_normal_types(data_list)

    if dump_mode == DumpMode.OVERWRITE:
        dump_mode = 'w'
    elif dump_mode == DumpMode.APPEND:
        dump_mode = 'a'
    with jsonl.open(jsonl_file, mode=dump_mode) as writer:
        writer.write_all(data_list)


def jsonl_to_csv(jsonl_file, csv_file):
    """
    Convert jsonl file to csv file.

    Args:
        jsonl_file: jsonl file path.
        csv_file: csv file path.
    """
    data = jsonl_to_list(jsonl_file)
    if not data:
        logger.warning(f'No data found in {jsonl_file}.')
        return

    with open(csv_file, 'w', newline='', encoding='utf-8') as f:
        writer = csv.writer(f)
        writer.writerow(data[0].keys())  # Write header
        for item in data:
            writer.writerow(item.values())


def csv_to_list(csv_file) -> list:
    """
    Read csv file to list.

    Args:
        csv_file: csv file path.

    Returns:
        list: list of lines. Each line is a dict.
    """
    res_list = []
    with open(csv_file, 'r', encoding='utf-8') as f:
        reader = csv.DictReader(f)
        for row in reader:
            res_list.append(row)
    return res_list


def tsv_to_list(tsv_file) -> list:
    """
    Read tsv file to list.

    Args:
        tsv_file: tsv file path.

    Returns:
        list: list of lines. Each line is a dict.
    """
    res_list = []
    with open(tsv_file, 'r', encoding='utf-8') as f:
        reader = csv.DictReader(f, delimiter='\t')
        for row in reader:
            res_list.append(row)
    return res_list


def csv_to_jsonl(csv_file, jsonl_file):
    """
    Convert csv file to jsonl file.

    Args:
        csv_file: csv file path.
        jsonl_file: jsonl file path.
    """
    data = csv_to_list(csv_file)
    if not data:
        logger.warning(f'No data found in {csv_file}.')
        return

    dump_jsonl_data(data, jsonl_file, dump_mode=DumpMode.OVERWRITE)


def yaml_to_dict(yaml_file) -> dict:
    """
    Read yaml file to dict.
    """
    with open(yaml_file, 'r') as f:
        try:
            stream = yaml.safe_load(f)
        except yaml.YAMLError as e:
            logger.error(f'{e}')
            raise e

    return stream


def dict_to_yaml(d: dict, yaml_file: str):
    """
    Dump dict to yaml file.
    """
    with open(yaml_file, 'w') as f:
        yaml.dump(d, f, default_flow_style=False, allow_unicode=True)


def json_to_dict(json_file) -> dict:
    """
    Read json file to dict.
    """
    with open(json_file, 'r') as f:
        try:
            stream = json.load(f)
        except json.JSONDecodeError as e:
            logger.error(f'{e}')
            raise e

    return stream


def are_paths_same(path1, path2):
    """
    Check if two paths are the same.
    """
    real_path1 = os.path.realpath(os.path.abspath(os.path.expanduser(path1)))
    real_path2 = os.path.realpath(os.path.abspath(os.path.expanduser(path2)))

    return real_path1 == real_path2


def dict_to_json(d: dict, json_file: str):
    """
    Dump dict to json file.
    """
    with open(json_file, 'w') as f:
        json.dump(d, f, indent=4, ensure_ascii=False)


def get_latest_folder_path(work_dir):
    from datetime import datetime

    # Get all subdirectories in the work_dir
    folders = [f for f in os.listdir(work_dir) if os.path.isdir(os.path.join(work_dir, f))]

    # Get the timestamp（YYYYMMDD_HHMMSS）
    timestamp_pattern = re.compile(r'^\d{8}_\d{6}$')

    # Filter out the folders
    timestamped_folders = [f for f in folders if timestamp_pattern.match(f)]

    if not timestamped_folders:
        print(f'>> No timestamped folders found in {work_dir}!')
        return None

    # timestamp parser
    def parse_timestamp(folder_name):
        return datetime.strptime(folder_name, '%Y%m%d_%H%M%S')

    # Find the latest folder
    latest_folder = max(timestamped_folders, key=parse_timestamp)

    return os.path.join(work_dir, latest_folder)


def gen_hash(name: str, bits: int = 32):
    return hashlib.md5(name.encode(encoding='UTF-8')).hexdigest()[:bits]


def get_valid_list(input_list, candidate_list):
    """
    Get the valid and invalid list from input_list based on candidate_list.
    Args:
        input_list: The input list.
        candidate_list: The candidate list.

    Returns:
        valid_list: The valid list.
        invalid_list: The invalid list.
    """
    return [i for i in input_list if i in candidate_list], \
           [i for i in input_list if i not in candidate_list]


def PIL_to_base64(image: Image.Image, format: str = 'JPEG', add_header: bool = False) -> str:
    """
    Convert a PIL Image to a base64 encoded string.

    Args:
        image (Image.Image): The PIL Image to convert.
        format (str): The format to save the image in. Default is 'JPEG'.
        add_header (bool): Whether to add the base64 header. Default is False.

    Returns:
        str: Base64 encoded string of the image.
    """
    buffered = BytesIO()
    image.save(buffered, format=format)
    img_str = base64.b64encode(buffered.getvalue()).decode('utf-8')
    if add_header:
        img_str = f'data:image/{format.lower()};base64,{img_str}'
    return img_str


def bytes_to_base64(bytes_data: bytes, *, format: str = 'png', add_header: bool = False, content_type='image') -> str:
    """Convert bytes to a base64 encoded string.

    Args:
        bytes_data (bytes): The bytes to convert.
        format (str): The format of the image. Default is 'png'.
        add_header (bool): Whether to add the base64 header. Default is False.
        content_type (str): The type of the data, 'image' or 'audio'. Default is 'image'.

    Returns:
        str: Base64 encoded string of the bytes.
    """
    base64_str = base64.b64encode(bytes_data).decode('utf-8')
    if add_header:
        base64_str = f'data:{content_type}/{format};base64,{base64_str}'
    return base64_str


def base64_to_PIL(base64_str):
    """Convert a base64 encoded string to a PIL Image.

    Args:
        base64_str (str): The base64 encoded string.

    Returns:
        Image.Image: The decoded PIL Image.
    """
    # remove header
    if ',' in base64_str:
        base64_str = base64_str.split(',', 1)[1]

    # decode
    img_data = base64.b64decode(base64_str)
    img_file = io.BytesIO(img_data)
    img = Image.open(img_file)
    return img


def safe_filename(s: str, max_length: int = 255) -> str:
    """
    Convert a string into a safe filename by removing or replacing unsafe characters.

    Args:
        s (str): The input string to convert
        max_length (int): Maximum length of the resulting filename (default 255)

    Returns:
        str: A safe filename string

    Examples:
        >>> safe_filename("Hello/World?.txt")
        'Hello_World.txt'
    """
    # normalize unicode characters
    s = unicodedata.normalize('NFKD', s)
    s = s.encode('ASCII', 'ignore').decode('ASCII')

    # remove or replace unsafe characters
    # Keep only alphanumeric characters, dots, dashes, and underscores
    safe_chars = string.ascii_letters + string.digits + '.-_'
    s = ''.join(c if c in safe_chars else '_' for c in s)

    # remove consecutive underscores
    s = re.sub(r'_+', '_', s)

    # remove leading/trailing periods and underscores
    s = s.strip('._')

    # handle empty string case
    if not s:
        s = 'untitled'

    # handle starting with a period (hidden files)
    if s.startswith('.'):
        s = '_' + s

    # enforce length limit
    if len(s) > max_length:
        # If we need to truncate, preserve the file extension if present
        name, ext = os.path.splitext(s)
        ext_len = len(ext)
        if ext_len > 0:
            max_name_length = max_length - ext_len
            s = name[:max_name_length] + ext
        else:
            s = s[:max_length]

    return s


def convert_normal_types(obj):
    """Recursively convert numpy types and datetime objects to native Python types for JSON serialization."""
    import numpy as np

    if isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, np.bool_):
        return bool(obj)
    elif isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, dict):
        return {key: convert_normal_types(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [convert_normal_types(item) for item in obj]
    elif isinstance(obj, tuple):
        return tuple(convert_normal_types(item) for item in obj)
    else:
        return obj


def compress_image_to_limit(image_bytes: bytes, max_bytes: int = 10_000_000) -> Tuple[bytes, str]:
    """
    Ensure image bytes are under max_bytes by re-encoding to JPEG with quality reduction
    and optional downscaling. Returns (processed_bytes, format_str).
    If the original bytes are already below the limit, returns them as PNG.
    """
    if len(image_bytes) <= max_bytes:
        return image_bytes, 'png'

    try:
        img = Image.open(BytesIO(image_bytes))
    except Exception as exc:
        logger.warning(f'Failed to open image bytes with PIL, sending original image; may exceed API limit: {exc}')
        return image_bytes, 'png'

    # Convert to RGB for JPEG if needed
    if img.mode not in ('RGB', 'L'):
        img = img.convert('RGB')

    def encode_jpeg(source: Image.Image, quality: int) -> bytes:
        buf = BytesIO()
        source.save(buf, format='JPEG', quality=quality, optimize=True, progressive=True)
        return buf.getvalue()

    # Start with moderate quality and reduce
    quality: int = 85
    out: bytes = encode_jpeg(img, quality)
    quality_floor: int = 40

    while len(out) > max_bytes and quality > quality_floor:
        quality -= 10
        out = encode_jpeg(img, quality)

    # If still too large, progressively downscale
    min_side_floor: int = 256
    scale: float = 0.9
    while len(out) > max_bytes and min(img.size) > min_side_floor:
        new_w = max(min_side_floor, int(img.width * scale))
        new_h = max(min_side_floor, int(img.height * scale))
        if (new_w, new_h) == img.size:
            break
        img = img.resize((new_w, new_h), Image.LANCZOS)
        out = encode_jpeg(img, quality)

    if len(out) > max_bytes:
        logger.warning(f'Image remains above limit after compression: size={len(out)} bytes (limit={max_bytes}).')
    else:
        logger.info(
            f'Compressed image from {len(image_bytes)} to {len(out)} bytes; '
            f'quality={quality}, size={img.width}x{img.height}.'
        )

    return out, 'jpeg'



import asyncio
import copy
import os
import platform
import threading
import time
import warnings
from argparse import Namespace
from logging import warn

from evalscope.perf.utils.local_server import start_app
from evalscope.perf.utils.log_utils import init_swanlab, init_wandb
from evalscope.utils.logger import configure_logging, get_logger
from evalscope.utils.model_utils import seed_everything
from .arguments import Arguments, parse_args
from .benchmark import benchmark
from .utils.db_util import get_output_path
from .utils.handler import add_signal_handlers
from .utils.rich_display import print_summary

logger = get_logger()


def run_one_benchmark(args: Arguments, output_path: str = None):
    if isinstance(args.parallel, list):
        args.parallel = args.parallel[0]
    if isinstance(args.number, list):
        args.number = args.number[0]

    # Setup logger and output
    args.outputs_dir = output_path

    logger.info('Starting benchmark with args: ')
    logger.info(args)

    if platform.system() == 'Windows':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    loop = asyncio.new_event_loop()
    if platform.system() != 'Windows':
        add_signal_handlers(loop)

    return loop.run_until_complete(benchmark(args))


def run_multi_benchmark(args: Arguments, output_path: str = None):
    results = []
    number_list = copy.deepcopy(args.number)
    parallel_list = copy.deepcopy(args.parallel)
    for i, (number, parallel) in enumerate(zip(number_list, parallel_list)):
        args.number = number
        args.parallel = parallel
        # Set up output path for each run
        cur_output_path = os.path.join(output_path, f'parallel_{parallel}_number_{number}')
        os.makedirs(cur_output_path, exist_ok=True)
        # Start the benchmark
        metrics_result = run_one_benchmark(args, output_path=cur_output_path)
        # Save the results
        results.append(metrics_result)
        # Sleep between runs to avoid overwhelming the server
        if i < len(number_list) - 1:
            logger.info(f'Sleeping for {args.sleep_interval} seconds before the next run...')
            time.sleep(args.sleep_interval)
    # Analyze results
    print_summary(results, args.model_id)
    return results


def run_perf_benchmark(args):
    # Check if args is a dictionary or Namespace
    if isinstance(args, dict):
        args = Arguments(**args)
    elif isinstance(args, Namespace):
        args = Arguments.from_args(args)

    if args.seed is not None:
        seed_everything(args.seed)

    # Initialize output directory
    output_path = get_output_path(args)
    configure_logging(args.debug, os.path.join(output_path, 'benchmark.log'))

    # Initialize wandb and swanlab
    visualizer = args.visualizer
    if visualizer is None:
        if args.wandb_api_key is not None:
            visualizer = 'wandb'
            warnings.warn('--wandb-api-key is deprecated. Please use `--visualizer wandb` instead.', DeprecationWarning)
        elif args.swanlab_api_key is not None:
            visualizer = 'swanlab'
            warnings.warn(
                '--swanlab-api-key is deprecated. Please use `--visualizer swanlab` instead.', DeprecationWarning
            )
    args.visualizer = visualizer
    if visualizer == 'wandb':
        init_wandb(args)
    elif visualizer == 'swanlab':
        init_swanlab(args)

    # Initialize local server if needed
    if args.api.startswith('local'):
        #  start local server
        server = threading.Thread(target=start_app, args=(copy.deepcopy(args), ), daemon=True)
        server.start()
    # Start benchmark
    if len(args.number) == 1:
        return run_one_benchmark(args, output_path=output_path)
    else:
        return run_multi_benchmark(args, output_path=output_path)


if __name__ == '__main__':
    args = Arguments.from_args(parse_args())
    metrics_result, percentile_result = run_perf_benchmark(args)
    print(metrics_result)
    print(percentile_result)



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import Any, Dict, List, Optional, Union

from evalscope.api.benchmark import BenchmarkMeta, DefaultDataAdapter
from evalscope.api.dataset import Sample
from evalscope.api.evaluator import TaskState
from evalscope.api.messages import ChatMessageSystem, ChatMessageUser, dict_to_chat_message
from evalscope.api.metric import Score
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.logger import get_logger

logger = get_logger()

PROMPT_TEMPLATE = '请回答问题\n{question}'


@register_benchmark(
    BenchmarkMeta(
        name='general_qa',
        pretty_name='General-QA',
        description='A general question answering dataset for custom evaluation. '
        'For detailed instructions on how to use this benchmark, please refer to the [User Guide](https://evalscope.readthedocs.io/en/latest/advanced_guides/custom_dataset/llm.html#qa).',  # noqa: E501
        tags=[Tags.QA, Tags.CUSTOM],
        dataset_id='general_qa',
        metric_list=['BLEU', 'Rouge'],
        few_shot_num=0,
        train_split=None,
        eval_split='test',
        prompt_template=PROMPT_TEMPLATE,
    )
)
class GeneralQAAdapter(DefaultDataAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def load_from_disk(self, **kwargs):
        return super().load_from_disk(use_local_loader=True)

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        """
        Convert a data record to a Sample object.

        Args:
            record (Dict[str, Any]): Input data record.

        Returns:
            Sample: Sample object with input, target, and metadata.
        """
        query = record.get('question') or record.get('query')
        answer = record.get('answer') or record.get('response')
        system_prompt = record.get('system')
        messages = record.get('messages')

        message_list = []
        if messages:
            message_list = [dict_to_chat_message(m) for m in messages]
        else:
            if system_prompt:
                message_list.append(ChatMessageSystem(content=system_prompt))
            message_list.append(ChatMessageUser(content=query))

        return Sample(input=message_list, target=answer or '')

    def match_score(
        self, original_prediction: str, filtered_prediction: str, reference: str, task_state: TaskState
    ) -> Score:
        """
        Calculate evaluation scores by comparing prediction with reference.
        """
        # Initialize the score object with prediction details
        score = Score(
            extracted_prediction=filtered_prediction,
            prediction=original_prediction,
        )

        # Calculate scores for each configured metric
        for metric in self.metric_list:
            try:
                if metric == 'Rouge':
                    from evalscope.metrics.rouge_metric import compute_rouge_score_one_sample_zh

                    score.value.update(compute_rouge_score_one_sample_zh([filtered_prediction], [reference]))
                elif metric == 'BLEU':
                    from evalscope.metrics import bleu_ngram_one_sample

                    score.value.update(bleu_ngram_one_sample(filtered_prediction, reference))
            except Exception as e:
                logger.error(f'Error calculating metric {metric}: {e}')
                return None

        score.main_score_name = 'Rouge-L-R'
        return score



from .default_data_adapter import DefaultDataAdapter
from .image_edit_adapter import ImageEditAdapter
from .multi_choice_adapter import MultiChoiceAdapter
from .ner_adapter import NERAdapter
from .text2image_adapter import Text2ImageAdapter
from .vision_language_adapter import VisionLanguageAdapter



from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, VisionLanguageAdapter
from evalscope.api.dataset import Sample
from evalscope.api.evaluator import TaskState
from evalscope.api.messages import ChatMessageUser, Content, ContentImage, ContentText
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.io_utils import bytes_to_base64
from evalscope.utils.logger import get_logger
from evalscope.utils.multi_choices import MultipleChoiceTemplate, parse_answers, prompt

logger = get_logger()

SUBSET_LIST = ['default']

MULT_CHOICE_PROMPT = MultipleChoiceTemplate.SINGLE_ANSWER_COT


@register_benchmark(
    BenchmarkMeta(
        name='ai2d',
        pretty_name='AI2D',
        tags=[Tags.MULTI_MODAL, Tags.KNOWLEDGE, Tags.QA],
        description=
        'AI2D is a benchmark dataset for researching the understanding of diagrams by AI. It contains over 5,000 diverse diagrams from science textbooks (e.g., the water cycle, food webs). Each diagram is accompanied by multiple-choice questions that test an AI\'s ability to interpret visual elements, text labels, and their relationships. The benchmark is challenging because it requires jointly understanding the layout, symbols, and text to answer questions correctly.',  # noqa: E501
        dataset_id='lmms-lab/ai2d',
        subset_list=SUBSET_LIST,
        metric_list=['acc'],
        eval_split='test',
        prompt_template=MULT_CHOICE_PROMPT,
    )
)
class Ai2dAdapter(VisionLanguageAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        answers_list: list[str] = record['options']
        input_text = prompt(question=record['question'], choices=answers_list, template=self.prompt_template)
        content_list: list[Content] = [ContentText(text=input_text)]
        image = record.get('image')
        if image:
            image_base64 = bytes_to_base64(image['bytes'], format='png', add_header=True)
            content_list.append(ContentImage(image=image_base64))

        label_answer = chr(int(record['answer']) + ord('A'))

        return Sample(input=[ChatMessageUser(content=content_list)], choices=answers_list, target=label_answer)

    def extract_answer(self, prediction: str, task_state: TaskState) -> str:
        answers = parse_answers(task_state)
        return ''.join(sorted(list(answers)))



from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

from evalscope.utils.logger import get_logger

logger = get_logger()

Value = Dict[str, Union[int, float, bool]]


class Score(BaseModel):
    """Score generated by a scorer."""

    value: Value = Field(default_factory=dict)
    """Score value as a dictionary. Key is the score name, value is the score value.
    The first key is considered the main score by default."""

    extracted_prediction: Optional[str] = Field(default=None)
    """Answer extracted from model output (optional)"""

    prediction: Optional[str] = Field(default=None)
    """Original prediction text from the model (optional)"""

    explanation: Optional[str] = Field(default=None)
    """Explanation of score (optional)."""

    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict)
    """Additional metadata related to the score"""

    main_score_name: Optional[str] = Field(default=None)
    """Main score name, if applicable. This is used to indicate which score is the primary score in a multi-score scenario."""  # noqa: E501

    @property
    def main_value(self) -> Union[int, float, bool]:
        """Main score value."""
        if self.main_score_name and self.main_score_name in self.value:
            return self.value[self.main_score_name]
        elif self.value:
            # If main_score_name is not set or not found, use the first value and update main_score_name
            first_key = next(iter(self.value))
            self.main_score_name = first_key
            return self.value[first_key]
        return None

    @main_value.setter
    def main_value(self, value: Union[int, float, bool]):
        """Set the main score value."""
        if self.main_score_name:
            # If main_score_name is already set, use it
            self.value[self.main_score_name] = value
        elif self.value:
            # If no main_score_name but value dict exists, use the first key
            first_key = next(iter(self.value))
            self.main_score_name = first_key
            self.value[first_key] = value
        else:
            # If neither main_score_name nor value dict exists, initialize both
            self.main_score_name = 'default'
            self.value[self.main_score_name] = value


class SampleScore(BaseModel):
    """Score for a Sample."""

    score: Score
    """A score"""

    sample_id: Optional[Union[str, int]] = Field(default=None)
    """A sample id"""

    group_id: Optional[Union[str, int]] = Field(default=None)
    """A group id for the sample, used for grouping k repeated samples."""

    sample_metadata: Optional[Dict[str, Any]] = Field(default=None)
    """Metadata from the sample"""


class AggScore(BaseModel):
    """Output of an aggregation operation."""

    score: float = Field(default=0.0)
    """Aggregated value as a float."""

    metric_name: str = Field(default='')
    """Name of the metric being aggregated."""

    aggregation_name: str = Field(default='')
    """Name of the aggregation methods"""

    num: int = Field(default=0)
    """Number of samples used in the aggregation."""

    ids: Optional[List[Union[str, int]]] = Field(default=None)
    """List of sample IDs used in the aggregation, if applicable."""

    metadata: Optional[Dict[str, Any]] = Field(default=None)
    """Additional metadata related to the aggregation."""


class Aggregator:

    name = 'default'

    def __call__(self, scores: List[SampleScore]) -> List[AggScore]:
        r"""Aggregate a metric on a list of scores.

        Args:
          scores: List of scores.

        Returns:
          List[AggregatOutput]: List of aggregated outputs.
        """
        ...



import torch
from modelscope import AutoProcessor
from transformers import CLIPConfig
from typing import List

from ...constants import CACHE_DIR
from ..model import ScoreModel

MPS_MODELS = ['mps']


class MPSModel(ScoreModel):
    'A wrapper for MPS Score models'

    def __init__(self, model_name='mps', device='cuda', cache_dir=CACHE_DIR):
        assert model_name in MPS_MODELS
        super().__init__(model_name=model_name, device=device, cache_dir=cache_dir)

    def load_model(self):
        """Load the model, tokenizer, image transform
        """
        from ..utils import download_file
        from .build_mps_model.clip_model import CLIPModel

        assert self.model_name == 'mps'

        processor_name_or_path = 'laion/CLIP-ViT-H-14-laion2B-s32B-b79K'
        self.processor = AutoProcessor.from_pretrained(processor_name_or_path)

        config = download_file('AI-ModelScope/MPS', file_name='config.json', cache_dir=self.cache_dir)
        model_pretrained_path = download_file(
            'AI-ModelScope/MPS', file_name='MPS_overall_state_dict.pt', cache_dir=self.cache_dir
        )  # modelscope model
        model_weight = torch.load(model_pretrained_path, weights_only=True, map_location='cpu')

        self.model = CLIPModel(config=CLIPConfig.from_json_file(config))
        self.model.load_state_dict(model_weight, strict=False)
        self.model.eval().to(self.device)

    def load_images(self, image: List[str]) -> torch.Tensor:
        """Load the image(s), and return a tensor (no preprocessing!!) put on self.device
        """
        image = [self.image_loader(x) for x in image]
        image = self.processor(images=image, return_tensors='pt')['pixel_values']
        return image

    def process_text(self, text: List[str]) -> dict:
        """Process the text(s), and return a tensor (after preprocessing) put on self.device
        """
        text_inputs = self.processor(
            text=text,
            padding='max_length',
            truncation=True,
            return_tensors='pt',
        ).input_ids
        return text_inputs

    @torch.no_grad()
    def forward(self, images: List[str], texts: List[str], condition=None) -> torch.Tensor:
        """Forward pass of the model to return n scores for n (image, text) pairs (in PyTorch Tensor)
        """
        assert len(images) == len(texts)
        image_input = self.load_images(images).to(self.device)
        text_input = self.process_text(texts).to(self.device)
        if condition is None:
            condition = 'light, color, clarity, tone, style, ambiance, artistry, shape, face, hair, hands, limbs, structure, instance, texture, quantity, attributes, position, number, location, word, things.'
        condition_batch = self.process_text(condition).repeat(text_input.shape[0], 1).to(self.device)

        # embed
        text_f, text_features = self.model.model.get_text_features(text_input)

        image_f = self.model.model.get_image_features(image_input.half())
        condition_f, _ = self.model.model.get_text_features(condition_batch)

        sim_text_condition = torch.einsum('b i d, b j d -> b j i', text_f, condition_f)
        sim_text_condition = torch.max(sim_text_condition, dim=1, keepdim=True)[0]
        sim_text_condition = sim_text_condition / sim_text_condition.max()
        mask = torch.where(sim_text_condition > 0.3, 0, float('-inf'))
        mask = mask.repeat(1, image_f.shape[1], 1)
        image_features = self.model.cross_model(image_f, text_f, mask.half())[:, 0, :]

        image_features = image_features / image_features.norm(dim=-1, keepdim=True)
        text_features = text_features / text_features.norm(dim=-1, keepdim=True)
        image_score = self.model.logit_scale.exp() * text_features @ image_features.T

        return image_score[0]



import base64
import httpx
import mimetypes
import re


def is_http_url(url: str) -> bool:
    return url.startswith('http://') or url.startswith('https://')


def is_data_uri(url: str) -> bool:
    pattern = r'^data:([^;]+);base64,.*'
    return re.match(pattern, url) is not None


def data_uri_mime_type(data_url: str) -> str | None:
    pattern = r'^data:([^;]+);.*'
    match = re.match(pattern, data_url)
    if match:
        mime_type = match.group(1)
        return mime_type
    else:
        return None


def data_uri_to_base64(data_uri: str) -> str:
    pattern = r'^data:[^,]+,'
    stripped_uri = re.sub(pattern, '', data_uri)
    return stripped_uri


def file_as_data(file: str) -> tuple[bytes, str]:
    if is_data_uri(file):
        # resolve mime type and base64 content
        mime_type = data_uri_mime_type(file) or 'image/png'
        file_base64 = data_uri_to_base64(file)
        file_bytes = base64.b64decode(file_base64)
    else:
        # guess mime type; need strict=False for webp images
        type, _ = mimetypes.guess_type(file, strict=False)
        if type:
            mime_type = type
        else:
            mime_type = 'image/png'

        # handle url or file
        if is_http_url(file):
            client = httpx.Client()
            file_bytes = client.get(file).content
        else:
            with open(file, 'rb') as f:
                file_bytes = f.read()

    # return bytes and type
    return file_bytes, mime_type


def file_as_data_uri(file: str) -> str:
    if is_data_uri(file):
        return file
    else:
        bytes, mime_type = file_as_data(file)
        base64_file = base64.b64encode(bytes).decode('utf-8')
        file = f'data:{mime_type};base64,{base64_file}'
        return file



from ...constants import CACHE_DIR
from .blip2_itm_model import BLIP2_ITM_MODELS, BLIP2ITMScoreModel
from .fga_blip2_model import FGA_BLIP2_MODELS, FGA_BLIP2ScoreModel
from .image_reward_model import IMAGE_REWARD_MODELS, ImageRewardScoreModel

ALL_ITM_MODELS = [
    BLIP2_ITM_MODELS,
    IMAGE_REWARD_MODELS,
    FGA_BLIP2_MODELS,
]


def list_all_itmscore_models():
    return [model for models in ALL_ITM_MODELS for model in models]


def get_itmscore_model(model_name, device='cuda', cache_dir=CACHE_DIR):
    assert model_name in list_all_itmscore_models()
    if model_name in BLIP2_ITM_MODELS:
        return BLIP2ITMScoreModel(model_name, device=device, cache_dir=cache_dir)
    elif model_name in IMAGE_REWARD_MODELS:
        return ImageRewardScoreModel(model_name, device=device, cache_dir=cache_dir)
    elif model_name in FGA_BLIP2_MODELS:
        return FGA_BLIP2ScoreModel(model_name, device=device, cache_dir=cache_dir)
    else:
        raise NotImplementedError()



import json
import yaml
from typing import Any, Dict, List, Optional

from evalscope.utils import get_logger
from .tool_call import ToolCall, ToolFunction
from .tool_info import ToolInfo

logger = get_logger()


def parse_tool_call(id: str, function: str, arguments: str, tools: Optional[List[ToolInfo]] = None) -> ToolCall:
    """Parse a tool call from a JSON payload.

    Note that this function doesn't know about internal tool names so the caller
    should ammend the returned `ToolCall` by mapping the parsed `function` field from
    from an internal name to an inspect tool name and fixing up the `ToolCall` object
    as required to reflect this change.
    """
    error: Optional[str] = None
    arguments_dict: Dict[str, Any] = {}

    def report_parse_error(ex: Exception) -> None:
        nonlocal error
        error = tool_parse_error_message(arguments, ex)
        logger.info(error)

    # if the arguments is a dict, then handle it with a plain json.loads
    arguments = arguments.strip()
    if arguments.startswith('{'):
        try:
            arguments_dict = json.loads(arguments)
        except json.JSONDecodeError as ex:
            report_parse_error(ex)

    # otherwise parse it as yaml (which will pickup unquoted strings, numbers, and true/false)
    # and then create a dict that maps it to the first function argument
    elif function and tools:
        tool_info = next(
            (tool for tool in tools if tool.name == function and len(tool.parameters.properties) > 0),
            None,
        )
        if tool_info:
            param_names = list(tool_info.parameters.properties.keys())
            try:
                value = yaml.safe_load(arguments)
                arguments_dict[param_names[0]] = value
            except yaml.error.YAMLError:
                # If the yaml parser fails, we treat it as a string argument.
                arguments_dict[param_names[0]] = arguments

    # return ToolCall with error payload
    return ToolCall(
        id=id,
        function=ToolFunction(
            name=function,
            arguments=arguments_dict,
        ),
        parse_error=error,
    )


def tool_parse_error_message(arguments: str, ex: Exception) -> str:
    return f'Error parsing the following tool call arguments:\n\n{arguments}\n\nError details: {ex}'



from .base import DatasetPluginBase
from .custom import CustomDatasetPlugin
from .flickr8k import FlickrDatasetPlugin
from .kontext_bench import KontextDatasetPlugin
from .line_by_line import LineByLineDatasetPlugin
from .longalpaca import LongAlpacaDatasetPlugin
from .openqa import OpenqaDatasetPlugin
from .random_dataset import RandomDatasetPlugin
from .random_vl_dataset import RandomVLDatasetPlugin
from .speed_benchmark import SpeedBenchmarkDatasetPlugin, SpeedBenchmarkLongDatasetPlugin



from evalscope.api.benchmark import BenchmarkMeta, NERAdapter
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.ner import FEWSHOT_TEMPLATE, PROMPT_TEMPLATE


@register_benchmark(
    BenchmarkMeta(
        name='conll2003',
        pretty_name='CoNLL2003',
        dataset_id='evalscope/conll2003',
        tags=[Tags.KNOWLEDGE, Tags.NER],
        description='The ConLL-2003 dataset is for the Named Entity Recognition (NER) task. It was introduced as part '
        'of the ConLL-2003 Shared Task conference and contains texts annotated with entities such as '
        'people, organizations, places, and various names.',
        few_shot_num=5,
        train_split='train',
        eval_split='test',
        metric_list=['precision', 'recall', 'f1_score', 'accuracy'],
        prompt_template=PROMPT_TEMPLATE,
        few_shot_prompt_template=FEWSHOT_TEMPLATE,
    )
)
class CoNLL2003Adapter(NERAdapter):
    """
    Adapter for the CoNLL2003 Named Entity Recognition dataset.

    This adapter inherits the NER functionality from NERAdapter and
    configures it specifically for the CoNLL2003 dataset's entity types.
    """

    def __init__(self, **kwargs):
        # Initialize the parent class first
        super().__init__(**kwargs)

        # Define CoNLL2003-specific entity mappings
        self.entity_type_map = {'PER': 'person', 'ORG': 'organization', 'LOC': 'location', 'MISC': 'miscellaneous'}

        # Add descriptions for each entity type
        self.entity_descriptions = {
            'PER': 'Names of people, including first and last names',
            'ORG': 'Names of companies, institutions, organizations, etc.',
            'LOC': 'Names of locations, cities, states, countries, etc.',
            'MISC': 'Miscellaneous entities not in the above categories'
        }

        # Setup entity mappings based on the defined entity types
        self.setup_entity_mappings()



from typing import Dict, Iterator, List

from evalscope.perf.arguments import Arguments
from evalscope.perf.plugin.datasets.base import DatasetPluginBase
from evalscope.perf.plugin.registry import register_dataset


@register_dataset('custom')
class CustomDatasetPlugin(DatasetPluginBase):
    """Read dataset and return prompt.
    """

    def __init__(self, query_parameters: Arguments):
        super().__init__(query_parameters)

    def build_messages(self) -> Iterator[List[Dict]]:
        for item in self.dataset_line_by_line(self.query_parameters.dataset_path):
            prompt = item.strip()
            is_valid, _ = self.check_prompt_length(prompt)
            if is_valid:
                if self.query_parameters.apply_chat_template:
                    message = self.create_message(prompt)
                    yield [message]
                else:
                    yield prompt


if __name__ == '__main__':
    from evalscope.perf.arguments import Arguments
    from evalscope.perf.main import run_perf_benchmark

    args = Arguments(
        model='qwen2.5-7b-instruct',
        url='https://dashscope.aliyuncs.com/compatible-mode/v1/chat/completions',
        dataset_path='outputs/perf_data.txt',
        api_key='EMPTY',
        dataset='custom',
    )

    run_perf_benchmark(args)



# Copyright 2023 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility library of instructions."""

import functools
import nltk
import os
import random
import re

RANK = os.environ.get('LOCAL_RANK', '0')

WORD_LIST = [
    'western',
    'sentence',
    'signal',
    'dump',
    'spot',
    'opposite',
    'bottom',
    'potato',
    'administration',
    'working',
    'welcome',
    'morning',
    'good',
    'agency',
    'primary',
    'wish',
    'responsibility',
    'press',
    'problem',
    'president',
    'steal',
    'brush',
    'read',
    'type',
    'beat',
    'trainer',
    'growth',
    'lock',
    'bone',
    'case',
    'equal',
    'comfortable',
    'region',
    'replacement',
    'performance',
    'mate',
    'walk',
    'medicine',
    'film',
    'thing',
    'rock',
    'tap',
    'total',
    'competition',
    'ease',
    'south',
    'establishment',
    'gather',
    'parking',
    'world',
    'plenty',
    'breath',
    'claim',
    'alcohol',
    'trade',
    'dear',
    'highlight',
    'street',
    'matter',
    'decision',
    'mess',
    'agreement',
    'studio',
    'coach',
    'assist',
    'brain',
    'wing',
    'style',
    'private',
    'top',
    'brown',
    'leg',
    'buy',
    'procedure',
    'method',
    'speed',
    'high',
    'company',
    'valuable',
    'pie',
    'analyst',
    'session',
    'pattern',
    'district',
    'pleasure',
    'dinner',
    'swimming',
    'joke',
    'order',
    'plate',
    'department',
    'motor',
    'cell',
    'spend',
    'cabinet',
    'difference',
    'power',
    'examination',
    'engine',
    'horse',
    'dimension',
    'pay',
    'toe',
    'curve',
    'literature',
    'bother',
    'fire',
    'possibility',
    'debate',
    'activity',
    'passage',
    'hello',
    'cycle',
    'background',
    'quiet',
    'author',
    'effect',
    'actor',
    'page',
    'bicycle',
    'error',
    'throat',
    'attack',
    'character',
    'phone',
    'tea',
    'increase',
    'outcome',
    'file',
    'specific',
    'inspector',
    'internal',
    'potential',
    'staff',
    'building',
    'employer',
    'shoe',
    'hand',
    'direction',
    'garden',
    'purchase',
    'interview',
    'study',
    'recognition',
    'member',
    'spiritual',
    'oven',
    'sandwich',
    'weird',
    'passenger',
    'particular',
    'response',
    'reaction',
    'size',
    'variation',
    'a',
    'cancel',
    'candy',
    'exit',
    'guest',
    'condition',
    'fly',
    'price',
    'weakness',
    'convert',
    'hotel',
    'great',
    'mouth',
    'mind',
    'song',
    'sugar',
    'suspect',
    'telephone',
    'ear',
    'roof',
    'paint',
    'refrigerator',
    'organization',
    'jury',
    'reward',
    'engineering',
    'day',
    'possession',
    'crew',
    'bar',
    'road',
    'description',
    'celebration',
    'score',
    'mark',
    'letter',
    'shower',
    'suggestion',
    'sir',
    'luck',
    'national',
    'progress',
    'hall',
    'stroke',
    'theory',
    'offer',
    'story',
    'tax',
    'definition',
    'history',
    'ride',
    'medium',
    'opening',
    'glass',
    'elevator',
    'stomach',
    'question',
    'ability',
    'leading',
    'village',
    'computer',
    'city',
    'grand',
    'confidence',
    'candle',
    'priest',
    'recommendation',
    'point',
    'necessary',
    'body',
    'desk',
    'secret',
    'horror',
    'noise',
    'culture',
    'warning',
    'water',
    'round',
    'diet',
    'flower',
    'bus',
    'tough',
    'permission',
    'week',
    'prompt',
    'connection',
    'abuse',
    'height',
    'save',
    'corner',
    'border',
    'stress',
    'drive',
    'stop',
    'rip',
    'meal',
    'listen',
    'confusion',
    'girlfriend',
    'living',
    'relation',
    'significance',
    'plan',
    'creative',
    'atmosphere',
    'blame',
    'invite',
    'housing',
    'paper',
    'drink',
    'roll',
    'silver',
    'drunk',
    'age',
    'damage',
    'smoke',
    'environment',
    'pack',
    'savings',
    'influence',
    'tourist',
    'rain',
    'post',
    'sign',
    'grandmother',
    'run',
    'profit',
    'push',
    'clerk',
    'final',
    'wine',
    'swim',
    'pause',
    'stuff',
    'singer',
    'funeral',
    'average',
    'source',
    'scene',
    'tradition',
    'personal',
    'snow',
    'nobody',
    'distance',
    'sort',
    'sensitive',
    'animal',
    'major',
    'negotiation',
    'click',
    'mood',
    'period',
    'arrival',
    'expression',
    'holiday',
    'repeat',
    'dust',
    'closet',
    'gold',
    'bad',
    'sail',
    'combination',
    'clothes',
    'emphasis',
    'duty',
    'black',
    'step',
    'school',
    'jump',
    'document',
    'professional',
    'lip',
    'chemical',
    'front',
    'wake',
    'while',
    'inside',
    'watch',
    'row',
    'subject',
    'penalty',
    'balance',
    'possible',
    'adult',
    'aside',
    'sample',
    'appeal',
    'wedding',
    'depth',
    'king',
    'award',
    'wife',
    'blow',
    'site',
    'camp',
    'music',
    'safe',
    'gift',
    'fault',
    'guess',
    'act',
    'shame',
    'drama',
    'capital',
    'exam',
    'stupid',
    'record',
    'sound',
    'swing',
    'novel',
    'minimum',
    'ratio',
    'machine',
    'shape',
    'lead',
    'operation',
    'salary',
    'cloud',
    'affair',
    'hit',
    'chapter',
    'stage',
    'quantity',
    'access',
    'army',
    'chain',
    'traffic',
    'kick',
    'analysis',
    'airport',
    'time',
    'vacation',
    'philosophy',
    'ball',
    'chest',
    'thanks',
    'place',
    'mountain',
    'advertising',
    'red',
    'past',
    'rent',
    'return',
    'tour',
    'house',
    'construction',
    'net',
    'native',
    'war',
    'figure',
    'fee',
    'spray',
    'user',
    'dirt',
    'shot',
    'task',
    'stick',
    'friend',
    'software',
    'promotion',
    'interaction',
    'surround',
    'block',
    'purpose',
    'practice',
    'conflict',
    'routine',
    'requirement',
    'bonus',
    'hole',
    'state',
    'junior',
    'sweet',
    'catch',
    'tear',
    'fold',
    'wall',
    'editor',
    'life',
    'position',
    'pound',
    'respect',
    'bathroom',
    'coat',
    'script',
    'job',
    'teach',
    'birth',
    'view',
    'resolve',
    'theme',
    'employee',
    'doubt',
    'market',
    'education',
    'serve',
    'recover',
    'tone',
    'harm',
    'miss',
    'union',
    'understanding',
    'cow',
    'river',
    'association',
    'concept',
    'training',
    'recipe',
    'relationship',
    'reserve',
    'depression',
    'proof',
    'hair',
    'revenue',
    'independent',
    'lift',
    'assignment',
    'temporary',
    'amount',
    'loss',
    'edge',
    'track',
    'check',
    'rope',
    'estimate',
    'pollution',
    'stable',
    'message',
    'delivery',
    'perspective',
    'mirror',
    'assistant',
    'representative',
    'witness',
    'nature',
    'judge',
    'fruit',
    'tip',
    'devil',
    'town',
    'emergency',
    'upper',
    'drop',
    'stay',
    'human',
    'neck',
    'speaker',
    'network',
    'sing',
    'resist',
    'league',
    'trip',
    'signature',
    'lawyer',
    'importance',
    'gas',
    'choice',
    'engineer',
    'success',
    'part',
    'external',
    'worker',
    'simple',
    'quarter',
    'student',
    'heart',
    'pass',
    'spite',
    'shift',
    'rough',
    'lady',
    'grass',
    'community',
    'garage',
    'youth',
    'standard',
    'skirt',
    'promise',
    'blind',
    'television',
    'disease',
    'commission',
    'positive',
    'energy',
    'calm',
    'presence',
    'tune',
    'basis',
    'preference',
    'head',
    'common',
    'cut',
    'somewhere',
    'presentation',
    'current',
    'thought',
    'revolution',
    'effort',
    'master',
    'implement',
    'republic',
    'floor',
    'principle',
    'stranger',
    'shoulder',
    'grade',
    'button',
    'tennis',
    'police',
    'collection',
    'account',
    'register',
    'glove',
    'divide',
    'professor',
    'chair',
    'priority',
    'combine',
    'peace',
    'extension',
    'maybe',
    'evening',
    'frame',
    'sister',
    'wave',
    'code',
    'application',
    'mouse',
    'match',
    'counter',
    'bottle',
    'half',
    'cheek',
    'resolution',
    'back',
    'knowledge',
    'make',
    'discussion',
    'screw',
    'length',
    'accident',
    'battle',
    'dress',
    'knee',
    'log',
    'package',
    'it',
    'turn',
    'hearing',
    'newspaper',
    'layer',
    'wealth',
    'profile',
    'imagination',
    'answer',
    'weekend',
    'teacher',
    'appearance',
    'meet',
    'bike',
    'rise',
    'belt',
    'crash',
    'bowl',
    'equivalent',
    'support',
    'image',
    'poem',
    'risk',
    'excitement',
    'remote',
    'secretary',
    'public',
    'produce',
    'plane',
    'display',
    'money',
    'sand',
    'situation',
    'punch',
    'customer',
    'title',
    'shake',
    'mortgage',
    'option',
    'number',
    'pop',
    'window',
    'extent',
    'nothing',
    'experience',
    'opinion',
    'departure',
    'dance',
    'indication',
    'boy',
    'material',
    'band',
    'leader',
    'sun',
    'beautiful',
    'muscle',
    'farmer',
    'variety',
    'fat',
    'handle',
    'director',
    'opportunity',
    'calendar',
    'outside',
    'pace',
    'bath',
    'fish',
    'consequence',
    'put',
    'owner',
    'go',
    'doctor',
    'information',
    'share',
    'hurt',
    'protection',
    'career',
    'finance',
    'force',
    'golf',
    'garbage',
    'aspect',
    'kid',
    'food',
    'boot',
    'milk',
    'respond',
    'objective',
    'reality',
    'raw',
    'ring',
    'mall',
    'one',
    'impact',
    'area',
    'news',
    'international',
    'series',
    'impress',
    'mother',
    'shelter',
    'strike',
    'loan',
    'month',
    'seat',
    'anything',
    'entertainment',
    'familiar',
    'clue',
    'year',
    'glad',
    'supermarket',
    'natural',
    'god',
    'cost',
    'conversation',
    'tie',
    'ruin',
    'comfort',
    'earth',
    'storm',
    'percentage',
    'assistance',
    'budget',
    'strength',
    'beginning',
    'sleep',
    'other',
    'young',
    'unit',
    'fill',
    'store',
    'desire',
    'hide',
    'value',
    'cup',
    'maintenance',
    'nurse',
    'function',
    'tower',
    'role',
    'class',
    'camera',
    'database',
    'panic',
    'nation',
    'basket',
    'ice',
    'art',
    'spirit',
    'chart',
    'exchange',
    'feedback',
    'statement',
    'reputation',
    'search',
    'hunt',
    'exercise',
    'nasty',
    'notice',
    'male',
    'yard',
    'annual',
    'collar',
    'date',
    'platform',
    'plant',
    'fortune',
    'passion',
    'friendship',
    'spread',
    'cancer',
    'ticket',
    'attitude',
    'island',
    'active',
    'object',
    'service',
    'buyer',
    'bite',
    'card',
    'face',
    'steak',
    'proposal',
    'patient',
    'heat',
    'rule',
    'resident',
    'broad',
    'politics',
    'west',
    'knife',
    'expert',
    'girl',
    'design',
    'salt',
    'baseball',
    'grab',
    'inspection',
    'cousin',
    'couple',
    'magazine',
    'cook',
    'dependent',
    'security',
    'chicken',
    'version',
    'currency',
    'ladder',
    'scheme',
    'kitchen',
    'employment',
    'local',
    'attention',
    'manager',
    'fact',
    'cover',
    'sad',
    'guard',
    'relative',
    'county',
    'rate',
    'lunch',
    'program',
    'initiative',
    'gear',
    'bridge',
    'breast',
    'talk',
    'dish',
    'guarantee',
    'beer',
    'vehicle',
    'reception',
    'woman',
    'substance',
    'copy',
    'lecture',
    'advantage',
    'park',
    'cold',
    'death',
    'mix',
    'hold',
    'scale',
    'tomorrow',
    'blood',
    'request',
    'green',
    'cookie',
    'church',
    'strip',
    'forever',
    'beyond',
    'debt',
    'tackle',
    'wash',
    'following',
    'feel',
    'maximum',
    'sector',
    'sea',
    'property',
    'economics',
    'menu',
    'bench',
    'try',
    'language',
    'start',
    'call',
    'solid',
    'address',
    'income',
    'foot',
    'senior',
    'honey',
    'few',
    'mixture',
    'cash',
    'grocery',
    'link',
    'map',
    'form',
    'factor',
    'pot',
    'model',
    'writer',
    'farm',
    'winter',
    'skill',
    'anywhere',
    'birthday',
    'policy',
    'release',
    'husband',
    'lab',
    'hurry',
    'mail',
    'equipment',
    'sink',
    'pair',
    'driver',
    'consideration',
    'leather',
    'skin',
    'blue',
    'boat',
    'sale',
    'brick',
    'two',
    'feed',
    'square',
    'dot',
    'rush',
    'dream',
    'location',
    'afternoon',
    'manufacturer',
    'control',
    'occasion',
    'trouble',
    'introduction',
    'advice',
    'bet',
    'eat',
    'kill',
    'category',
    'manner',
    'office',
    'estate',
    'pride',
    'awareness',
    'slip',
    'crack',
    'client',
    'nail',
    'shoot',
    'membership',
    'soft',
    'anybody',
    'web',
    'official',
    'individual',
    'pizza',
    'interest',
    'bag',
    'spell',
    'profession',
    'queen',
    'deal',
    'resource',
    'ship',
    'guy',
    'chocolate',
    'joint',
    'formal',
    'upstairs',
    'car',
    'resort',
    'abroad',
    'dealer',
    'associate',
    'finger',
    'surgery',
    'comment',
    'team',
    'detail',
    'crazy',
    'path',
    'tale',
    'initial',
    'arm',
    'radio',
    'demand',
    'single',
    'draw',
    'yellow',
    'contest',
    'piece',
    'quote',
    'pull',
    'commercial',
    'shirt',
    'contribution',
    'cream',
    'channel',
    'suit',
    'discipline',
    'instruction',
    'concert',
    'speech',
    'low',
    'effective',
    'hang',
    'scratch',
    'industry',
    'breakfast',
    'lay',
    'join',
    'metal',
    'bedroom',
    'minute',
    'product',
    'rest',
    'temperature',
    'many',
    'give',
    'argument',
    'print',
    'purple',
    'laugh',
    'health',
    'credit',
    'investment',
    'sell',
    'setting',
    'lesson',
    'egg',
    'middle',
    'marriage',
    'level',
    'evidence',
    'phrase',
    'love',
    'self',
    'benefit',
    'guidance',
    'affect',
    'you',
    'dad',
    'anxiety',
    'special',
    'boyfriend',
    'test',
    'blank',
    'payment',
    'soup',
    'obligation',
    'reply',
    'smile',
    'deep',
    'complaint',
    'addition',
    'review',
    'box',
    'towel',
    'minor',
    'fun',
    'soil',
    'issue',
    'cigarette',
    'internet',
    'gain',
    'tell',
    'entry',
    'spare',
    'incident',
    'family',
    'refuse',
    'branch',
    'can',
    'pen',
    'grandfather',
    'constant',
    'tank',
    'uncle',
    'climate',
    'ground',
    'volume',
    'communication',
    'kind',
    'poet',
    'child',
    'screen',
    'mine',
    'quit',
    'gene',
    'lack',
    'charity',
    'memory',
    'tooth',
    'fear',
    'mention',
    'marketing',
    'reveal',
    'reason',
    'court',
    'season',
    'freedom',
    'land',
    'sport',
    'audience',
    'classroom',
    'law',
    'hook',
    'win',
    'carry',
    'eye',
    'smell',
    'distribution',
    'research',
    'country',
    'dare',
    'hope',
    'whereas',
    'stretch',
    'library',
    'if',
    'delay',
    'college',
    'plastic',
    'book',
    'present',
    'use',
    'worry',
    'champion',
    'goal',
    'economy',
    'march',
    'election',
    'reflection',
    'midnight',
    'slide',
    'inflation',
    'action',
    'challenge',
    'guitar',
    'coast',
    'apple',
    'campaign',
    'field',
    'jacket',
    'sense',
    'way',
    'visual',
    'remove',
    'weather',
    'trash',
    'cable',
    'regret',
    'buddy',
    'beach',
    'historian',
    'courage',
    'sympathy',
    'truck',
    'tension',
    'permit',
    'nose',
    'bed',
    'son',
    'person',
    'base',
    'meat',
    'usual',
    'air',
    'meeting',
    'worth',
    'game',
    'independence',
    'physical',
    'brief',
    'play',
    'raise',
    'board',
    'she',
    'key',
    'writing',
    'pick',
    'command',
    'party',
    'yesterday',
    'spring',
    'candidate',
    'physics',
    'university',
    'concern',
    'development',
    'change',
    'string',
    'target',
    'instance',
    'room',
    'bitter',
    'bird',
    'football',
    'normal',
    'split',
    'impression',
    'wood',
    'long',
    'meaning',
    'stock',
    'cap',
    'leadership',
    'media',
    'ambition',
    'fishing',
    'essay',
    'salad',
    'repair',
    'today',
    'designer',
    'night',
    'bank',
    'drawing',
    'inevitable',
    'phase',
    'vast',
    'chip',
    'anger',
    'switch',
    'cry',
    'twist',
    'personality',
    'attempt',
    'storage',
    'being',
    'preparation',
    'bat',
    'selection',
    'white',
    'technology',
    'contract',
    'side',
    'section',
    'station',
    'till',
    'structure',
    'tongue',
    'taste',
    'truth',
    'difficulty',
    'group',
    'limit',
    'main',
    'move',
    'feeling',
    'light',
    'example',
    'mission',
    'might',
    'wait',
    'wheel',
    'shop',
    'host',
    'classic',
    'alternative',
    'cause',
    'agent',
    'consist',
    'table',
    'airline',
    'text',
    'pool',
    'craft',
    'range',
    'fuel',
    'tool',
    'partner',
    'load',
    'entrance',
    'deposit',
    'hate',
    'article',
    'video',
    'summer',
    'feature',
    'extreme',
    'mobile',
    'hospital',
    'flight',
    'fall',
    'pension',
    'piano',
    'fail',
    'result',
    'rub',
    'gap',
    'system',
    'report',
    'suck',
    'ordinary',
    'wind',
    'nerve',
    'ask',
    'shine',
    'note',
    'line',
    'mom',
    'perception',
    'brother',
    'reference',
    'bend',
    'charge',
    'treat',
    'trick',
    'term',
    'homework',
    'bake',
    'bid',
    'status',
    'project',
    'strategy',
    'orange',
    'let',
    'enthusiasm',
    'parent',
    'concentrate',
    'device',
    'travel',
    'poetry',
    'business',
    'society',
    'kiss',
    'end',
    'vegetable',
    'employ',
    'schedule',
    'hour',
    'brave',
    'focus',
    'process',
    'movie',
    'illegal',
    'general',
    'coffee',
    'ad',
    'highway',
    'chemistry',
    'psychology',
    'hire',
    'bell',
    'conference',
    'relief',
    'show',
    'neat',
    'funny',
    'weight',
    'quality',
    'club',
    'daughter',
    'zone',
    'touch',
    'tonight',
    'shock',
    'burn',
    'excuse',
    'name',
    'survey',
    'landscape',
    'advance',
    'satisfaction',
    'bread',
    'disaster',
    'item',
    'hat',
    'prior',
    'shopping',
    'visit',
    'east',
    'photo',
    'home',
    'idea',
    'father',
    'comparison',
    'cat',
    'pipe',
    'winner',
    'count',
    'lake',
    'fight',
    'prize',
    'foundation',
    'dog',
    'keep',
    'ideal',
    'fan',
    'struggle',
    'peak',
    'safety',
    'solution',
    'hell',
    'conclusion',
    'population',
    'strain',
    'alarm',
    'measurement',
    'second',
    'train',
    'race',
    'due',
    'insurance',
    'boss',
    'tree',
    'monitor',
    'sick',
    'course',
    'drag',
    'appointment',
    'slice',
    'still',
    'care',
    'patience',
    'rich',
    'escape',
    'emotion',
    'royal',
    'female',
    'childhood',
    'government',
    'picture',
    'will',
    'sock',
    'big',
    'gate',
    'oil',
    'cross',
    'pin',
    'improvement',
    'championship',
    'silly',
    'help',
    'sky',
    'pitch',
    'man',
    'diamond',
    'most',
    'transition',
    'work',
    'science',
    'committee',
    'moment',
    'fix',
    'teaching',
    'dig',
    'specialist',
    'complex',
    'guide',
    'people',
    'dead',
    'voice',
    'original',
    'break',
    'topic',
    'data',
    'degree',
    'reading',
    'recording',
    'bunch',
    'reach',
    'judgment',
    'lie',
    'regular',
    'set',
    'painting',
    'mode',
    'list',
    'player',
    'bear',
    'north',
    'wonder',
    'carpet',
    'heavy',
    'officer',
    'negative',
    'clock',
    'unique',
    'baby',
    'pain',
    'assumption',
    'disk',
    'iron',
    'bill',
    'drawer',
    'look',
    'double',
    'mistake',
    'finish',
    'future',
    'brilliant',
    'contact',
    'math',
    'rice',
    'leave',
    'restaurant',
    'discount',
    'sex',
    'virus',
    'bit',
    'trust',
    'event',
    'wear',
    'juice',
    'failure',
    'bug',
    'context',
    'mud',
    'whole',
    'wrap',
    'intention',
    'draft',
    'pressure',
    'cake',
    'dark',
    'explanation',
    'space',
    'angle',
    'word',
    'efficiency',
    'management',
    'habit',
    'star',
    'chance',
    'finding',
    'transportation',
    'stand',
    'criticism',
    'flow',
    'door',
    'injury',
    'insect',
    'surprise',
    'apartment',
]  # pylint: disable=line-too-long

# ISO 639-1 codes to language names.
LANGUAGE_CODES = {
    'en': 'English',
    'es': 'Spanish',
    'pt': 'Portuguese',
    'ar': 'Arabic',
    'hi': 'Hindi',
    'fr': 'French',
    'ru': 'Russian',
    'de': 'German',
    'ja': 'Japanese',
    'it': 'Italian',
    'bn': 'Bengali',
    'uk': 'Ukrainian',
    'th': 'Thai',
    'ur': 'Urdu',
    'ta': 'Tamil',
    'te': 'Telugu',
    'bg': 'Bulgarian',
    'ko': 'Korean',
    'pl': 'Polish',
    'he': 'Hebrew',
    'fa': 'Persian',
    'vi': 'Vietnamese',
    'ne': 'Nepali',
    'sw': 'Swahili',
    'kn': 'Kannada',
    'mr': 'Marathi',
    'gu': 'Gujarati',
    'pa': 'Punjabi',
    'ml': 'Malayalam',
    'fi': 'Finnish',
}

_ALPHABETS = '([A-Za-z])'
_PREFIXES = '(Mr|St|Mrs|Ms|Dr)[.]'
_SUFFIXES = '(Inc|Ltd|Jr|Sr|Co)'
_STARTERS = r'(Mr|Mrs|Ms|Dr|Prof|Capt|Cpt|Lt|He\s|She\s|It\s|They\s|Their\s|Our\s|We\s|But\s|However\s|That\s|This\s|Wherever)'  # noqa: E501
_ACRONYMS = '([A-Z][.][A-Z][.](?:[A-Z][.])?)'
_WEBSITES = '[.](com|net|org|io|gov|edu|me)'
_DIGITS = '([0-9])'
_MULTIPLE_DOTS = r'\.{2,}'


def split_into_sentences(text):
    """Split the text into sentences.

    Args:
      text: A string that consists of more than or equal to one sentences.

    Returns:
      A list of strings where each string is a sentence.
    """
    text = ' ' + text + '  '
    text = text.replace('\n', ' ')
    text = re.sub(_PREFIXES, '\\1<prd>', text)
    text = re.sub(_WEBSITES, '<prd>\\1', text)
    text = re.sub(_DIGITS + '[.]' + _DIGITS, '\\1<prd>\\2', text)
    text = re.sub(
        _MULTIPLE_DOTS,
        lambda match: '<prd>' * len(match.group(0)) + '<stop>',
        text,
    )
    if 'Ph.D' in text:
        text = text.replace('Ph.D.', 'Ph<prd>D<prd>')
    text = re.sub(r'\s' + _ALPHABETS + '[.] ', ' \\1<prd> ', text)
    text = re.sub(_ACRONYMS + ' ' + _STARTERS, '\\1<stop> \\2', text)
    text = re.sub(
        _ALPHABETS + '[.]' + _ALPHABETS + '[.]' + _ALPHABETS + '[.]',
        '\\1<prd>\\2<prd>\\3<prd>',
        text,
    )
    text = re.sub(_ALPHABETS + '[.]' + _ALPHABETS + '[.]', '\\1<prd>\\2<prd>', text)
    text = re.sub(' ' + _SUFFIXES + '[.] ' + _STARTERS, ' \\1<stop> \\2', text)
    text = re.sub(' ' + _SUFFIXES + '[.]', ' \\1<prd>', text)
    text = re.sub(' ' + _ALPHABETS + '[.]', ' \\1<prd>', text)
    if '”' in text:
        text = text.replace('.”', '”.')
    if '"' in text:
        text = text.replace('."', '".')
    if '!' in text:
        text = text.replace('!"', '"!')
    if '?' in text:
        text = text.replace('?"', '"?')
    text = text.replace('.', '.<stop>')
    text = text.replace('?', '?<stop>')
    text = text.replace('!', '!<stop>')
    text = text.replace('<prd>', '.')
    sentences = text.split('<stop>')
    sentences = [s.strip() for s in sentences]
    if sentences and not sentences[-1]:
        sentences = sentences[:-1]
    return sentences


def count_words(text):
    """Counts the number of words."""
    tokenizer = nltk.tokenize.RegexpTokenizer(r'\w+')
    tokens = tokenizer.tokenize(text)
    num_words = len(tokens)
    return num_words


@functools.lru_cache(maxsize=None)
def _get_sentence_tokenizer():
    return nltk.data.load('nltk:tokenizers/punkt/english.pickle')


def count_sentences(text):
    """Count the number of sentences."""
    tokenizer = _get_sentence_tokenizer()
    tokenized_sentences = tokenizer.tokenize(text)
    return len(tokenized_sentences)


def generate_keywords(num_keywords):
    """Randomly generates a few keywords."""
    return random.sample(WORD_LIST, k=num_keywords)



from typing import Any, Dict, List, Set, Tuple

from evalscope.api.dataset import Sample
from evalscope.api.metric.scorer import AggScore, SampleScore, Score
from evalscope.utils.import_utils import check_import
from evalscope.utils.logger import get_logger
from evalscope.utils.ner import (
    DEFAULT_TAG_FIX_PATTERNS,
    calculate_bio_metrics,
    clean_prediction,
    create_target_text,
    extract_entities_from_text,
    extract_spans_from_bio,
    xml_to_bio_tags,
)
from .default_data_adapter import DefaultDataAdapter

logger = get_logger()


class NERAdapter(DefaultDataAdapter):
    """
    Base adapter class for Named Entity Recognition (NER) tasks.

    This adapter handles converting between BIO tagging schemes and XML-style entity markup,
    and provides evaluation metrics using seqeval.

    Subclasses should define their entity types and register the benchmark.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Define mapping from BIO tags to user-friendly tag names
        self.entity_type_map = {}
        # Add descriptions for each entity type
        self.entity_descriptions = {}

        # These will be initialized in setup_entity_mappings
        self.reverse_entity_map = {}
        self.entity_list = []
        self.entities_description = ''

        # Define common error patterns to handle
        self.tag_fix_patterns = DEFAULT_TAG_FIX_PATTERNS

        check_import('seqeval', 'seqeval', raise_error=True, feature_name='NER metrics')
        # Note: setup_entity_mappings() should be called by subclasses
        # after they define their entity_type_map and entity_descriptions

    def setup_entity_mappings(self):
        """
        Setup entity mappings and descriptions for prompt formatting.
        This should be called after entity_type_map and entity_descriptions are defined.
        """
        # Reverse mapping for converting back from prediction to evaluation
        self.reverse_entity_map = {v.lower(): k for k, v in self.entity_type_map.items()}

        # Create list of tags for prompt formatting
        self.entity_list = [f'<{ent.lower()}>' for ent in self.entity_type_map.values()]

        # Create description of entities for prompt
        self.entities_description = ', '.join([
            f'{self.entity_type_map[tag]} ({self.entity_descriptions[tag]})' for tag in self.entity_type_map
        ])

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        """
        Convert a record with tokens and NER tags into a Sample.
        Creates both the raw text input and annotated text target.
        """
        tokens: List[str] = record['tokens']
        ner_tags: List[str] = record['ner_tags']

        # Create the input text by joining tokens
        input_text = ' '.join(tokens)

        # Process tokens and tags to create annotated target text
        target_text = create_target_text(tokens, ner_tags, self.entity_type_map)

        # Store tokens and tags in metadata for evaluation
        metadata = {'tokens': tokens, 'ner_tags': ner_tags}

        return Sample(input=input_text, target=target_text, metadata=metadata)

    def format_prompt_template(self, sample):
        """
        Format the prompt with entity types, available tags, and text to annotate.
        """
        return self.prompt_template.format(
            entities=self.entities_description, entity_list=', '.join(self.entity_list), text=sample.input
        )

    def format_fewshot_template(self, fewshot, sample):
        """
        Format the few-shot prompt with all required parameters.
        """
        return self.few_shot_prompt_template.format(
            fewshot=fewshot,
            entities=self.entities_description,
            entity_list=', '.join(self.entity_list),
            text=sample.input
        )

    def sample_to_fewshot(self, sample: Sample) -> str:
        """
        Format a sample as a few-shot example showing original and annotated text.
        """
        if not sample.metadata:
            return ''

        # Format few-shot examples to match the expected response format
        return f'Input:\n{sample.input}\n\nOutput:\n{sample.target}'

    def match_score(self, original_prediction, filtered_prediction, reference, task_state) -> Score:
        """
        Evaluate named entity recognition performance using seqeval.
        """
        from seqeval.metrics import accuracy_score, f1_score, precision_score, recall_score

        score = Score(
            extracted_prediction=filtered_prediction,
            prediction=original_prediction,
        )

        try:
            # Get the original tokens and tags from the reference metadata
            original_tokens = task_state.metadata['tokens']
            original_tags = task_state.metadata['ner_tags']

            if not original_tokens or len(original_tokens) == 0:
                if hasattr(reference, 'metadata') and reference.metadata:
                    original_tokens = reference.metadata['tokens']
                    original_tags = reference.metadata['ner_tags']

            # Clean and normalize the prediction
            cleaned_prediction = clean_prediction(filtered_prediction, self.tag_fix_patterns)

            # Convert XML-style prediction back to BIO tags aligned with original tokens
            pred_bio_tags = xml_to_bio_tags(cleaned_prediction, original_tokens, self.reverse_entity_map)

            # Use seqeval to calculate metrics
            # Note: seqeval expects lists of lists (one per sequence)
            y_true = [original_tags]
            y_pred = [pred_bio_tags]

            precision = precision_score(y_true, y_pred)
            recall = recall_score(y_true, y_pred)
            f1 = f1_score(y_true, y_pred)
            accuracy = accuracy_score(y_true, y_pred)

            score.value = {'precision': precision, 'recall': recall, 'f1_score': f1, 'accuracy': accuracy}

            # Store tags for aggregation (proper micro-averaging in aggregate_scores)
            # This way aggregate_scores can compute metrics across all samples at once,
            # which gives you true micro-averaged scores rather than averaged macro scores.
            score.metadata = {'y_true': original_tags, 'y_pred': pred_bio_tags}
        except Exception as e:
            logger.warning(f'Error evaluating NER prediction: {str(e)}')
            score.value = {'precision': 0.0, 'recall': 0.0, 'f1_score': 0.0, 'accuracy': 0.0}

        return score

    def aggregate_scores(self, sample_scores: List[SampleScore]) -> List[AggScore]:
        """
        Aggregate metrics across all samples using seqeval.
        """
        from seqeval.metrics import accuracy_score, f1_score, precision_score, recall_score

        # Collect all predictions and references
        y_true_all = []
        y_pred_all = []

        for ss in sample_scores:
            # Extract the BIO tags from metadata if available
            # You may need to store these during match_score
            if hasattr(ss.score, 'metadata') and 'y_true' in ss.score.metadata and 'y_pred' in ss.score.metadata:
                y_true_all.append(ss.score.metadata['y_true'])
                y_pred_all.append(ss.score.metadata['y_pred'])

        if not y_true_all:
            # Fallback: calculate averages from individual scores
            num_samples = len(sample_scores)
            avg_precision = sum(ss.score.value.get('precision', 0.0) for ss in sample_scores) / num_samples
            avg_recall = sum(ss.score.value.get('recall', 0.0) for ss in sample_scores) / num_samples
            avg_f1 = sum(ss.score.value.get('f1_score', 0.0) for ss in sample_scores) / num_samples
            avg_accuracy = sum(ss.score.value.get('accuracy', 0.0) for ss in sample_scores) / num_samples
        else:
            # Use seqeval for micro-averaged metrics across all samples
            avg_precision = precision_score(y_true_all, y_pred_all)
            avg_recall = recall_score(y_true_all, y_pred_all)
            avg_f1 = f1_score(y_true_all, y_pred_all)
            avg_accuracy = accuracy_score(y_true_all, y_pred_all)

        num_samples = len(sample_scores)

        agg_scores = [
            AggScore(
                metric_name='precision',
                score=avg_precision,
                num=num_samples,
                metadata={'type': 'seqeval-micro-average'}
            ),
            AggScore(
                metric_name='recall', score=avg_recall, num=num_samples, metadata={'type': 'seqeval-micro-average'}
            ),
            AggScore(metric_name='f1_score', score=avg_f1, num=num_samples, metadata={'type': 'seqeval-micro-average'}),
            AggScore(
                metric_name='accuracy', score=avg_accuracy, num=num_samples, metadata={'type': 'seqeval-accuracy'}
            )
        ]

        return agg_scores



import copy
import os
import random
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

from evalscope.api.dataset.utils import record_to_sample_fn
from evalscope.constants import DEFAULT_EVALSCOPE_CACHE_DIR, HubType
from evalscope.utils import get_logger
from evalscope.utils.io_utils import csv_to_list, gen_hash, jsonl_to_list, safe_filename, tsv_to_list
from .dataset import Dataset, FieldSpec, MemoryDataset, Sample
from .utils import data_to_samples, shuffle_choices_if_requested

logger = get_logger()


class DataLoader(ABC):
    """
    Abstract base class for data loaders.
    """

    def __init__(
        self,
        data_id_or_path: str,
        split: str,
        sample_fields: Union[FieldSpec, Callable] = None,
        filter_func: Callable = None,
        subset: str = 'default',
        version: str = None,
        limit: Union[int, float] = None,
        data_source: Optional[str] = None,
        shuffle: bool = False,
        shuffle_choices: Optional[Union[bool, int]] = None,
        seed: Optional[int] = None,
        auto_id: bool = True,
        repeats: int = 1,
        trust_remote: bool = True,
        **kwargs
    ):
        self.data_id_or_path = data_id_or_path
        self.split = split
        self.sample_fields = sample_fields
        self.filter_func = filter_func
        self.subset = subset
        self.version = version
        self.limit = limit
        self.data_source = data_source
        self.shuffle = shuffle
        self.shuffle_choices = shuffle_choices
        self.seed = seed
        self.auto_id = auto_id
        self.repeats = repeats
        self.trust_remote = trust_remote
        self.kwargs = kwargs

    @abstractmethod
    def load(self) -> Dataset:
        """
        Load data from the source.
        """
        ...


class RemoteDataLoader(DataLoader):
    """
    Data loader for remote datasets: ModelScope or Huggingface.
    """

    def load(self) -> Dataset:
        import datasets
        from modelscope import MsDataset

        path = self.data_id_or_path
        # resolve data_to_sample function
        data_to_sample = record_to_sample_fn(self.sample_fields)
        # generate a unique cache dir for this dataset
        dataset_hash = gen_hash(f'{path}{self.split}{self.subset}{self.version}{self.kwargs}')
        datasets_cache_dir = os.path.join(DEFAULT_EVALSCOPE_CACHE_DIR, 'datasets')
        dataset_cache_dir = os.path.join(datasets_cache_dir, f'{safe_filename(path)}-{dataset_hash}')
        if os.path.exists(dataset_cache_dir):
            dataset = datasets.load_from_disk(dataset_cache_dir)
        else:
            logger.info(
                f'Loading dataset {path} from {self.data_source} > subset: {self.subset} > split: {self.split} ...'
            )
            if self.data_source == HubType.MODELSCOPE:
                dataset = MsDataset.load(
                    dataset_name=path,
                    split=self.split,
                    subset_name=self.subset,
                    version=self.version,
                    trust_remote_code=self.trust_remote,
                    **self.kwargs,
                )
                # convert to Huggingface dataset if necessary
                if not isinstance(dataset, datasets.Dataset):
                    dataset = dataset.to_hf_dataset()
            elif self.data_source in [HubType.HUGGINGFACE, HubType.LOCAL]:
                # remove dataset_infos.json file if exists, since datasets will occur an error if it exists.
                dataset_infos_path = os.path.join(path, 'dataset_infos.json')
                if os.path.exists(dataset_infos_path):
                    logger.info(f'Removing dataset_infos.json file at {dataset_infos_path} to avoid datasets errors.')
                    os.remove(dataset_infos_path)
                # load dataset from Huggingface or local path
                dataset = datasets.load_dataset(
                    path=path,
                    name=self.subset if self.subset != 'default' else None,
                    split=self.split,
                    revision=self.version,
                    trust_remote_code=self.trust_remote,
                    **self.kwargs,
                )

            # Only save to disk if not loading from local path
            if self.data_source != HubType.LOCAL:
                dataset.save_to_disk(dataset_cache_dir)

        # shuffle if requested
        if self.shuffle:
            dataset = dataset.shuffle(seed=self.seed)

        # limit if requested
        if self.limit:
            if isinstance(self.limit, float):
                self.limit = int(len(dataset) * self.limit)
            elif isinstance(self.limit, int) and self.limit < 0:
                raise ValueError('Limit must be a non-negative integer or a float between 0 and 1.')
            if len(dataset) > self.limit:
                dataset = dataset.select(range(self.limit))

        # convert to list
        dataset = dataset.to_list()

        # repeat k times
        if self.repeats > 1:
            dataset = [copy.deepcopy(item) for item in dataset for _ in range(self.repeats)]

        # return the dataset
        memory_dataset = MemoryDataset(
            samples=data_to_samples(data=dataset, data_to_sample=data_to_sample),
            name=Path(path).stem if Path(path).exists() else path,
            location=path,
        )

        # Apply filtering if a filter function is provided
        if self.filter_func is not None:
            memory_dataset = memory_dataset.filter(self.filter_func)

        # assign ids and group_ids if requested
        if self.auto_id:
            memory_dataset.reindex(group_size=self.repeats)

        shuffle_choices_if_requested(memory_dataset, self.shuffle_choices)

        return memory_dataset


class LocalDataLoader(DataLoader):
    """
    Data loader for local datasets. Reads from JSONL or CSV files.
    """

    def load(self):

        path = self.data_id_or_path
        data_to_sample = record_to_sample_fn(self.sample_fields)
        dataset = []

        # Check for JSONL or CSV files in the specified path
        for ext, loader in [
            ('.jsonl', jsonl_to_list),
            ('.csv', csv_to_list),
            ('.tsv', tsv_to_list),
        ]:
            # Check if the file exists with the given extension
            if os.path.isfile(path) and path.endswith(ext):
                file_paths = [path]
            else:
                file_paths = [
                    os.path.join(path, f'{self.subset}_{self.split}{ext}'),
                    os.path.join(path, f'{self.subset}{ext}')
                ]
            # If the file exists, load it
            for file_path in file_paths:
                if os.path.exists(file_path):
                    dataset = loader(file_path)
                    break  # Stop checking other extensions once a file is found

        # shuffle if requested
        if self.shuffle:
            random.shuffle(dataset, self.seed)

        # limit if requested
        if self.limit:
            if isinstance(self.limit, float):
                self.limit = int(len(dataset) * self.limit)
            elif isinstance(self.limit, int) and self.limit < 0:
                raise ValueError('Limit must be a non-negative integer or a float between 0 and 1.')
            dataset = dataset[:self.limit]

        # repeat k times
        if self.repeats > 1:
            dataset = [copy.deepcopy(item) for item in dataset for _ in range(self.repeats)]

        # return the dataset
        memory_dataset = MemoryDataset(
            samples=data_to_samples(data=dataset, data_to_sample=data_to_sample),
            name=Path(path).stem if Path(path).exists() else path,
            location=path,
        )

        # Apply filtering if a filter function is provided
        if self.filter_func is not None:
            memory_dataset = memory_dataset.filter(self.filter_func)

        # assign ids and group_ids if requested
        if self.auto_id:
            memory_dataset.reindex(group_size=self.repeats)

        shuffle_choices_if_requested(memory_dataset, self.shuffle_choices)

        return memory_dataset


class DictDataLoader(DataLoader):
    """Load dataset from a list of dictionaries."""

    def __init__(self, dict_list: list, **kwargs):
        super().__init__(data_id_or_path='', split='', **kwargs)
        self.dict_list = dict_list

    def load(self) -> Dataset:
        data_to_sample = record_to_sample_fn(self.sample_fields)
        dataset = self.dict_list

        # shuffle if requested
        if self.shuffle:
            random.shuffle(dataset, self.seed)

        # limit if requested
        if self.limit:
            if isinstance(self.limit, float):
                self.limit = int(len(dataset) * self.limit)
            elif isinstance(self.limit, int) and self.limit < 0:
                raise ValueError('Limit must be a non-negative integer or a float between 0 and 1.')
            dataset = dataset[:self.limit]

        # repeat k times
        if self.repeats > 1:
            dataset = [copy.deepcopy(item) for item in dataset for _ in range(self.repeats)]

        # return the dataset
        memory_dataset = MemoryDataset(samples=data_to_samples(data=dataset, data_to_sample=data_to_sample), )

        # Apply filtering if a filter function is provided
        if self.filter_func is not None:
            memory_dataset = memory_dataset.filter(self.filter_func)

        # assign ids and group_ids if requested
        if self.auto_id:
            memory_dataset.reindex(group_size=self.repeats)

        shuffle_choices_if_requested(memory_dataset, self.shuffle_choices)

        return memory_dataset



import json
from typing import TYPE_CHECKING, Dict, List, Tuple

from evalscope.utils.logger import get_logger

if TYPE_CHECKING:
    from evalscope.api.mixin.sandbox_mixin import SandboxMixin

logger = get_logger()


def evaluate_in_sandbox(
    adapter: 'SandboxMixin',
    code: str,
    evaluation_sample: str,
    timeout: int = 6,
    debug: bool = False
) -> Tuple[bool, Dict]:
    """
    Evaluate code in sandbox environment for Live Code Bench.

    Args:
        adapter: The adapter instance with sandbox capabilities
        code: The code to evaluate
        evaluation_sample: JSON string containing input/output test cases
        timeout: Timeout for execution
        debug: Whether to enable debug logging

    Returns:
        Tuple[bool, Dict]: (overall_pass, detailed_results)
    """
    try:
        # Parse the evaluation sample
        test_data = json.loads(evaluation_sample)
        inputs = test_data.get('inputs', [])
        outputs = test_data.get('outputs', [])
        fn_name = test_data.get('fn_name')

        if debug:
            logger.info(f'Evaluating code with {len(inputs)} test cases')
            logger.info(f'Function name: {fn_name}')

        # Determine if this is call-based or stdio-based
        if fn_name:
            # Call-based evaluation
            return _evaluate_call_based_in_sandbox(adapter, code, inputs, outputs, fn_name, timeout, debug)
        else:
            # Standard input/output evaluation
            return _evaluate_stdio_in_sandbox(adapter, code, inputs, outputs, timeout, debug)

    except Exception as e:
        if debug:
            logger.error(f'Sandbox evaluation error: {str(e)}')
        return False, {'error': str(e), 'total_tests': 0, 'passed_tests': 0}


def _evaluate_call_based_in_sandbox(
    adapter: 'SandboxMixin', code: str, inputs: list, outputs: list, fn_name: str, timeout: int, debug: bool
) -> Tuple[bool, Dict]:
    """Evaluate call-based problems in sandbox."""
    try:
        all_passed = True
        passed_count = 0
        failed_cases = []

        for i, (test_input, expected_output) in enumerate(zip(inputs, outputs)):
            # Prepare individual test code for each test case
            test_code = f"""
import json
import sys

# User's code
{code}

# Test execution for single test case
try:
    test_input = {repr(test_input)}
    expected_output = {repr(expected_output)}

    if 'class Solution' in '''{code}''':
        # LeetCode style
        solution = Solution()
        method = getattr(solution, '{fn_name}')
    else:
        # Function is directly available
        method = {fn_name}

    # Parse input if it's JSON string
    if isinstance(test_input, str):
        try:
            test_input = json.loads(test_input)
        except:
            pass  # Keep as string if not valid JSON

    # Call the method
    if isinstance(test_input, list):
        result = method(*test_input)
    else:
        result = method(test_input)

    # Parse expected output if it's JSON string
    if isinstance(expected_output, str):
        try:
            expected_output = json.loads(expected_output)
        except:
            pass  # Keep as string if not valid JSON

    # Convert tuple to list for comparison
    if isinstance(result, tuple):
        result = list(result)

    if result == expected_output:
        print("TEST_PASSED")
    else:
        print(f"TEST_FAILED: expected {{expected_output}}, got {{result}}")

except Exception as e:
    print(f"EXECUTION_ERROR: {{str(e)}}")
    import traceback
    traceback.print_exc()
"""

            # Execute in sandbox
            result = adapter.execute_code_in_sandbox(code=test_code, timeout=timeout, language='python')

            if debug:
                logger.info(f'Test case {i} execution result: {result}')

            # Check if execution was successful and test passed
            if result.get('status') == 'success':
                output = result.get('output', '')
                if 'TEST_PASSED' in output:
                    passed_count += 1
                elif 'TEST_FAILED:' in output:
                    # Extract failure details from output
                    for line in output.split('\n'):
                        if line.startswith('TEST_FAILED:'):
                            failed_cases.append(f"Test {i}: {line.replace('TEST_FAILED: ', '')}")
                            break
                    all_passed = False
                    break
                elif 'EXECUTION_ERROR:' in output:
                    # Extract error details
                    for line in output.split('\n'):
                        if line.startswith('EXECUTION_ERROR:'):
                            failed_cases.append(f'Test {i}: {line}')
                            break
                    all_passed = False
                    break
                else:
                    failed_cases.append(f'Test {i}: Unknown error in output. Result: {result}')
                    all_passed = False
                    break
            else:
                failed_cases.append(f'Test {i}: Sandbox execution failed - Result: {result}')
                all_passed = False
                break

        detailed_results = {'total_tests': len(inputs), 'passed_tests': passed_count, 'failed_cases': failed_cases}

        return all_passed, detailed_results

    except Exception as e:
        if debug:
            logger.error(f'Call-based evaluation error: {str(e)}')
        return False, {'error': str(e), 'total_tests': len(inputs), 'passed_tests': 0}


def _evaluate_stdio_in_sandbox(
    adapter: 'SandboxMixin', code: str, inputs: list, outputs: list, timeout: int, debug: bool
) -> Tuple[bool, Dict]:
    """Evaluate stdio-based problems in sandbox."""
    try:
        all_passed = True
        passed_count = 0
        failed_cases = []

        for i, (test_input, expected_output) in enumerate(zip(inputs, outputs)):
            test_code = f"""
import sys
from io import StringIO

# Redirect stdin
sys.stdin = StringIO('''{test_input}''')

# User's code
{code}
"""

            # Execute in sandbox
            result = adapter.execute_code_in_sandbox(code=test_code, timeout=timeout, language='python')

            if result.get('status') != 'success':
                if debug:
                    logger.error(f'Test case {i} execution failed: {result}')
                failed_cases.append(f'Test {i}: Execution error - Result: {result}')
                all_passed = False
                break

            # Compare output
            actual_output = result.get('output', '').strip()
            expected_output = expected_output.strip()

            if actual_output == expected_output:
                passed_count += 1
            else:
                if debug:
                    logger.info(f"Test case {i} failed: expected '{expected_output}', got '{actual_output}'")
                failed_cases.append(f"Test {i}: Expected '{expected_output}', got '{actual_output}'")
                all_passed = False
                break

        detailed_results = {'total_tests': len(inputs), 'passed_tests': passed_count, 'failed_cases': failed_cases}

        return all_passed, detailed_results

    except Exception as e:
        if debug:
            logger.error(f'Stdio evaluation error: {str(e)}')
        return False, {'error': str(e), 'total_tests': len(inputs), 'passed_tests': 0}



import json
from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, DefaultDataAdapter
from evalscope.api.dataset import Sample
from evalscope.api.evaluator import TaskState
from evalscope.api.messages import ChatMessageUser, messages_pretty_str
from evalscope.api.metric import Score
from evalscope.api.model import Model
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.import_utils import check_import
from evalscope.utils.logger import get_logger

logger = get_logger()

SUBSET_LIST = [
    'Chinese',
    'English',
    'German',
    'Italian',
    'Vietnamese',
    'Spanish',
    'Hindi',
    'Portuguese',
    'French',
    'Thai',
    'Russian',
]


@register_benchmark(
    BenchmarkMeta(
        name='multi_if',
        pretty_name='Multi-IF',
        description=
        'Multi-IF is a benchmark designed to evaluate the performance of LLM models\' capabilities in multi-turn instruction following within a multilingual environment.',  # noqa: E501
        tags=[Tags.INSTRUCTION_FOLLOWING, Tags.MULTI_LINGUAL, Tags.MULTI_TURN],
        dataset_id='facebook/Multi-IF',
        subset_list=SUBSET_LIST,
        metric_list=[
            'prompt_level_strict',
            'inst_level_strict',
            'prompt_level_loose',
            'inst_level_loose',
        ],
        few_shot_num=0,
        train_split=None,
        eval_split='train',
        extra_params={
            'max_turns': 3,  # maximum number of turns to evaluate
        }
    )
)
class MultiIFAdapter(DefaultDataAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Ensure required packages are installed
        check_import(
            module_name=['nltk', 'langdetect'],
            package=['nltk', 'langdetect'],
            raise_error=True,
            feature_name=self.pretty_name
        )
        if 'Chinese' in self.subset_list:
            check_import(module_name='emoji', package='emoji', raise_error=True, feature_name='Chinese subset')
        if 'Thai' in self.subset_list:
            check_import(module_name='pythainlp', package='pythainlp', raise_error=True, feature_name='Thai subset')

        self.reformat_subset = True
        self.max_turns = self.extra_params.get('max_turns', 3)
        if not isinstance(self.max_turns, int) or self.max_turns < 1 or self.max_turns > 3:
            logger.warning(f'max_turns should be an integer between 1 and 3, got {self.max_turns}, clamping to 3.')
            self.max_turns = 3

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        return Sample(
            input=[ChatMessageUser(content='')],  # NOTE: we will build the multi turn conversation in the evaluator
            target='',
            subset_key=record['language'],
            metadata=record,
        )

    def run_inference(self, model: Model, sample: Sample, output_dir: str, **kwargs) -> TaskState:
        """
        Run multi-turn inference with the model and sample.
        """
        record = sample.metadata
        history = []
        step_record = {}
        for step in range(1, self.max_turns + 1):
            current_prompt = json.loads(record[f'turn_{step}_prompt'])
            history.append(ChatMessageUser(content=current_prompt['content']))
            # Generate model output
            model_output = model.generate(input=history, tools=sample.tools)

            response = model_output.completion
            instruction_id_list = json.loads(record[f'turn_{step}_instruction_id_list'])
            kwargs_list = json.loads(record[f'turn_{step}_kwargs'])
            _kwargs = [json.loads(kwarg) for kwarg in kwargs_list]

            step_record[step] = {
                'prompt': messages_pretty_str(history),
                'response': response,
                'instruction_id_list': instruction_id_list,
                'kwargs': _kwargs
            }

            # Append model output to history for next turn
            history.append(model_output.message)

        sample.metadata['step_record'] = step_record
        return TaskState(
            model=model.name,
            sample=sample,
            messages=history,
            output=model_output,
            completed=True,
        )

    def match_score(
        self, original_prediction: str, filtered_prediction: str, reference: Dict, task_state: TaskState
    ) -> Score:
        """
        Calculate evaluation scores by comparing prediction with reference.
        """
        from .metrics import gen_acc_loose, gen_acc_strict, parse_result

        # Initialize the score object with prediction details
        score = Score(
            extracted_prediction=filtered_prediction,
            prediction=original_prediction,
        )

        step_record = task_state.metadata['step_record']
        results = {}
        try:
            for step, record in step_record.items():
                outputs_strict = gen_acc_strict(record)
                outputs_loose = gen_acc_loose(record)
                prompt_level_strict, inst_level_strict = parse_result([outputs_strict])
                prompt_level_loose, inst_level_loose = parse_result([outputs_loose])
                results.update({
                    f'turn_{step}_prompt_level_strict': prompt_level_strict,
                    f'turn_{step}_inst_level_strict': inst_level_strict,
                    f'turn_{step}_prompt_level_loose': prompt_level_loose,
                    f'turn_{step}_inst_level_loose': inst_level_loose,
                })
            score.value.update(results)

            # Set main score name
            if results:
                score.main_score_name = f'turn_{step}_prompt_level_strict'

        except Exception as e:
            logger.error(f'Error calculating ifeval metrics: {e}')
            score.value = {}

        return score



# Copyright (c) Alibaba, Inc. and its affiliates.

from typing import TYPE_CHECKING

from .import_utils import _LazyModule

if TYPE_CHECKING:
    from .argument_utils import BaseArgument, get_supported_params, parse_int_or_float
    from .deprecation_utils import deprecated
    from .function_utils import run_once, thread_safe
    from .import_utils import get_module_path, is_module_installed
    from .io_utils import (
        OutputsStructure,
        csv_to_jsonl,
        csv_to_list,
        dict_to_yaml,
        gen_hash,
        get_latest_folder_path,
        get_valid_list,
        json_to_dict,
        jsonl_to_csv,
        jsonl_to_list,
        safe_filename,
        yaml_to_dict,
    )
    from .logger import configure_logging, get_logger
    from .model_utils import EvalBackend, dict_torch_dtype_to_str, fix_do_sample_warning, get_device, seed_everything

else:
    _import_structure = {
        'argument_utils': [
            'BaseArgument',
            'parse_int_or_float',
            'get_supported_params',
        ],
        'model_utils': [
            'EvalBackend',
            'get_device',
            'seed_everything',
            'dict_torch_dtype_to_str',
            'fix_do_sample_warning',
        ],
        'import_utils': [
            'is_module_installed',
            'get_module_path',
        ],
        'function_utils': [
            'thread_safe',
            'run_once',
        ],
        'io_utils': [
            'OutputsStructure',
            'csv_to_list',
            'json_to_dict',
            'yaml_to_dict',
            'get_latest_folder_path',
            'gen_hash',
            'dict_to_yaml',
            'csv_to_jsonl',
            'jsonl_to_csv',
            'jsonl_to_list',
            'gen_hash',
            'get_valid_list',
            'safe_filename',
            'thread_safe',
        ],
        'deprecation_utils': [
            'deprecated',
        ],
        'logger': [
            'get_logger',
            'configure_logging',
        ],
    }

    import sys

    sys.modules[__name__] = _LazyModule(
        __name__,
        globals()['__file__'],
        _import_structure,
        module_spec=__spec__,
        extra_objects={},
    )



import base64
import os
import torch
from openai import OpenAI
from typing import List

from .vqa_model import VQAScoreModel

default_question_template = 'Does this figure show "{}"? Please answer yes or no.'
default_answer_template = 'Yes'

GPT4V_MODELS = {
    # We recommend using 'gpt-4-turbo' for optimal performance.
    'gpt-4-turbo': {},
    'gpt-4o': {},
}


# Function to encode the image
def encode_image(image_path):
    with open(image_path, 'rb') as image_file:
        return base64.b64encode(image_file.read()).decode('utf-8')


def get_image_type(image_path):
    image_type = image_path.split('.')[-1]
    assert image_type in ['png', 'jpeg', 'jpg', 'gif', 'bmp', 'webp']
    return image_type


class GPT4VModel(VQAScoreModel):
    """A wrapper for the GPT4V models"""

    def __init__(self, model_name='gpt-4-turbo', device='cuda', cache_dir=None, openai_key=None, top_logprobs=2):
        assert model_name in GPT4V_MODELS
        assert openai_key is not None, 'Please provide an OpenAI API key'
        self.openai_key = openai_key
        self.top_logprobs = top_logprobs
        super().__init__(model_name=model_name, device=device, cache_dir=cache_dir)

    def load_model(self):
        """Load the model, tokenizer, image transform
        """
        import tiktoken

        self.tokenizer = tiktoken.encoding_for_model(self.model_name)
        self.client = OpenAI(api_key=self.openai_key)
        # self.candidate_answers = GPT4V_MODELS[self.model_name]['candidate_answers']
        # assert GPT4V_MODELS[self.model_name]['answer'] in self.candidate_answers
        # self.candidate_tokens = []
        # for ans in self.candidate_answers:
        #     token = self.tokenizer.encode(ans)
        #     assert len(token) == 1, "Currently only support single token answers"
        #     self.candidate_tokens.append(token[0])

    def load_images(self, image: List[str]) -> torch.Tensor:
        """Load the image(s), and return the string
        """
        image = [{'path': img, 'type': get_image_type(img), 'base64': encode_image(img)} for img in image]
        return image

    def forward_single(self, image, question, answer):
        try:
            completion = self.client.chat.completions.create(
                model=self.model_name,
                messages=[{
                    'role':
                    'user',
                    'content': [{
                        'type': 'text',
                        'text': question
                    }, {
                        'type': 'image_url',
                        'image_url': {
                            'url': f"data:image/{image['type']};base64,{image['base64']}"
                        }
                    }]
                }],
                logprobs=True,
                top_logprobs=self.top_logprobs,
                # logit_bias={yes_token:50, no_token:50}
            )
        except:
            print(
                f"Warning: completion not generated for image: {image['path']} and question: {question} and answer: {answer}"
            )
            print(f'Trying again with the same image')
            try:
                completion = self.client.chat.completions.create(
                    model=self.model_name,
                    messages=[{
                        'role':
                        'user',
                        'content': [{
                            'type': 'text',
                            'text': question
                        }, {
                            'type': 'image_url',
                            'image_url': {
                                'url': f"data:image/{image['type']};base64,{image['base64']}"
                            }
                        }]
                    }],
                    logprobs=True,
                    top_logprobs=self.top_logprobs,
                )
            except:
                print(f"Failed image: {image['path']} and question: {question} and answer: {answer}")
                return torch.Tensor([0.0])

        # print(completion.choices[0].message)
        # print(completion.choices[0].logprobs)
        # print(completion.choices[0].logprobs.content[0])
        is_generated = False
        for top_logprob in completion.choices[0].logprobs.content[0].top_logprobs:
            if top_logprob.token == answer:
                is_generated = True
                return torch.Tensor([top_logprob.logprob]).exp()
        if not is_generated:
            print(
                f"Warning: answer not generated for image: {image['path']} and question: {question} and answer: {answer}"
            )
            print(completion.choices[0].logprobs.content[0].top_logprobs)
            return torch.Tensor([0.0])

    def forward(
        self,
        images: List[str],
        texts: List[str],
        question_template: str = default_question_template,
        answer_template: str = default_answer_template
    ) -> torch.Tensor:
        """Forward pass of the model to return n scores for n (image, text) pairs (in PyTorch Tensor)
        """
        assert len(images) == len(texts), 'Number of images and texts must match'
        # Turn "a photo of a dog" into
        # Q: "Does this figure show "a photo of a dog"? Please answer yes or no."
        # A: "Yes"
        questions = [question_template.format(text) for text in texts]
        answers = [answer_template.format(text) for text in texts]

        for ans in answers:
            ans_tokens = self.tokenizer.encode(ans)
            assert len(ans_tokens) == 1, 'Currently only support single token answers'

        images = self.load_images(images)

        lm_prob = torch.zeros(len(images))

        for idx, (image, question, answer) in enumerate(zip(images, questions, answers)):
            lm_prob[idx] = self.forward_single(image, question, answer)

        return lm_prob



def get_entity_mappings():
    entity_type_map = {
        'COUNTRY': 'country',
        'ELECTION': 'election',
        'EVENT': 'event',
        'LOCATION': 'location',
        'MISC': 'misc',
        'ORGANISATION': 'organisation',
        'PERSON': 'person',
        'POLITICALPARTY': 'political_party',
        'POLITICIAN': 'politician'
    }
    entity_descriptions = {
        'COUNTRY': ('A country or sovereign state (e.g., "United States", "Germany").'),
        'ELECTION': ('A specific election event (e.g., "2024 presidential election", '
                     '"midterm elections").'),
        'EVENT':
        ('A significant political event, summit, or incident (e.g., "G7 Summit", '
         '"Brexit", "Watergate scandal").'),
        'LOCATION':
        ('A politically significant building or location (e.g., "The White House", '
         '"10 Downing Street").'),
        'MISC': (
            'Miscellaneous political terms, ideologies, or documents (e.g., "democracy", '
            '"impeachment", "the Constitution").'
        ),
        'ORGANISATION':
        ('A political or governmental organization (e.g., "United Nations", "NATO", '
         '"European Union").'),
        'PERSON':
        ('A person mentioned in a political context who is not a politician '
         '(e.g., a journalist, an activist).'),
        'POLITICALPARTY': ('A named political party (e.g., "Democratic Party", "Conservative Party").'),
        'POLITICIAN': ('A person who holds or seeks political office (e.g., "Joe Biden", '
                       '"Angela Merkel").')
    }
    return entity_type_map, entity_descriptions



# flake8: noqa: E501
from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, DefaultDataAdapter
from evalscope.api.dataset import Sample
from evalscope.api.evaluator import TaskState
from evalscope.api.metric import AggScore, SampleScore, Score
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.logger import get_logger

logger = get_logger()

GRADER_SYSTEM_PROMPT = """Please act as an impartial judge and evaluate the quality of the responses provided by two AI assistants to the user prompt displayed below. You will be given assistant A's answer and assistant B's answer. Your job is to evaluate which assistant's answer is better.\n\nBegin your evaluation by generating your own answer to the prompt. You must provide your answers before judging any answers.\n\nWhen evaluating the assistants' answers, compare both assistants' answers with your answer. You must identify and correct any mistakes or inaccurate information.\n\nThen consider if the assistant's answers are helpful, relevant, and concise. Helpful means the answer correctly responds to the prompt or follows the instructions. Note when user prompt has any ambiguity or more than one interpretation, it is more helpful and appropriate to ask for clarifications or more information from the user than providing an answer based on assumptions. Relevant means all parts of the response closely connect or are appropriate to what is being asked. Concise means the response is clear and not verbose or excessive.\n\nThen consider the creativity and novelty of the assistant's answers when needed. Finally, identify any missing important information in the assistants' answers that would be beneficial to include when responding to the user prompt.\n\nAfter providing your explanation, you must output only one of the following choices as your final verdict with a label:\n\n1. Assistant A is significantly better: [[A>>B]]\n2. Assistant A is slightly better: [[A>B]]\n3. Tie, relatively the same: [[A=B]]\n4. Assistant B is slightly better: [[B>A]]\n5. Assistant B is significantly better: [[B>>A]]\n\nExample output: \"My final verdict is tie: [[A=B]]\"."""  # noqa: E501

GRADER_TEMPLATE = """<|User Prompt|>\n{question}\n\n<|The Start of Assistant A's Answer|>\n{answer_1}\n<|The End of Assistant A's Answer|>\n\n<|The Start of Assistant B's Answer|>\n{answer_2}\n<|The End of Assistant B's Answer|>""".strip(
)


@register_benchmark(
    BenchmarkMeta(
        name='arena_hard',
        pretty_name='ArenaHard',
        tags=[Tags.INSTRUCTION_FOLLOWING, Tags.ARENA],
        description=
        'ArenaHard is a benchmark designed to evaluate the performance of large language models in a competitive setting, '
        'where models are pitted against each other in a series of tasks to determine their relative strengths and weaknesses. '
        'It includes a set of challenging tasks that require reasoning, understanding, and generation capabilities. '
        'Currently not support `style-controlled winrate`; the official Judge model is `gpt-4-1106-preview`, while the baseline model is `gpt-4-0314`.',
        dataset_id='AI-ModelScope/arena-hard-auto-v0.1',
        metric_list=['winrate'],
        few_shot_num=0,
        train_split=None,
        eval_split='test',
        prompt_template='{question}'
    )
)
class ArenaHardAdapter(DefaultDataAdapter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._use_llm_judge = True  # Use LLM as a judge by default

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        """
        Convert a data record to a Sample object.

        Args:
            record (Dict[str, Any]): Input data record.

        Returns:
            Sample: Sample object with input, target, and metadata.
        """
        question = record['question']
        baseline_prediction = record['prediction']  # baseline model prediction

        return Sample(
            input=question, target=baseline_prediction, metadata={'capability': record.get('capability', 'unknown')}
        )

    def llm_match_score(
        self,
        original_prediction: str,
        filtered_prediction: str,
        reference: str,
        task_state: TaskState,
    ) -> Score:
        from .utils import get_judge_score, post_process_arenahard

        score = Score(
            extracted_prediction=filtered_prediction,
            prediction=original_prediction,
        )

        question = task_state.input_text

        # reference is baseline answer 'A', filtered_prediction is model answer 'B'
        prompt1 = GRADER_TEMPLATE.format(question=question, answer_1=reference, answer_2=filtered_prediction)
        # reverse the order
        prompt2 = GRADER_TEMPLATE.format(question=question, answer_1=filtered_prediction, answer_2=reference)

        # get grading response
        game1_response = self.llm_judge.judge(prompt1, system_prompt=GRADER_SYSTEM_PROMPT)
        game2_response = self.llm_judge.judge(prompt2, system_prompt=GRADER_SYSTEM_PROMPT)

        # parse grading response
        res1 = post_process_arenahard(game1_response)
        res2 = post_process_arenahard(game2_response)

        score1 = get_judge_score(res1, reverse=True)
        score2 = get_judge_score(res2, reverse=False)

        battle_result = {
            'model_a':
            'gpt4-0314',
            'model_b':
            'test_model',
            'games': [
                {
                    'user_prompt': prompt1,
                    'judgment': game1_response,
                    'score': res1
                },
                {
                    'user_prompt': prompt2,
                    'judgment': game2_response,
                    'score': res2
                },
            ]
        }

        # Set score based on the battle result
        score.value = {'score': (score1 + score2) / 2}
        score.explanation = f'LLM judge battles: Game1: {game1_response[:100]}... Game2: {game2_response[:100]}...'
        score.metadata = {
            'source': 'llm_judge',
            'judge_strategy': self.judge_strategy,
            'model': self.llm_judge.model_id,
            'battle_result': battle_result
        }
        return score

    def aggregate_scores(self, sample_scores: List[SampleScore]) -> List[AggScore]:
        import pandas as pd

        from .utils import compute_mle_elo, get_battles_from_row, get_bootstrap_result, get_win_rate_column

        battles = pd.concat([get_battles_from_row(res.score.metadata['battle_result']) for res in sample_scores])

        bootstrap_online_elo = compute_mle_elo(battles)

        stats = pd.DataFrame()
        stats['results'] = None
        stats['results'] = stats['results'].astype('object')

        for i, model in enumerate(bootstrap_online_elo.index):
            # assert model in bootstrap_elo_lu.columns
            stats.at[i, 'model'] = model
            stats.at[i, 'score'] = bootstrap_online_elo[model]

        score = get_win_rate_column(stats, 'score', 'gpt4-0314').at['test_model']

        return [AggScore(
            score=score,
            metric_name='winrate',
            num=len(sample_scores),
        )]



# the following code is largely adapted from https://github.com/lework/llm-benchmark

import numpy as np
from rich.console import Console
from rich.panel import Panel
from rich.style import Style
from rich.table import Table
from rich.text import Text

from evalscope.utils.logger import get_logger
from .benchmark_util import Metrics
from .db_util import PercentileMetrics

logger = get_logger()


def analyze_results(all_results):
    """Analyze all test results and generate a summary report"""
    summary = []
    total_tokens = 0
    total_time = 0

    for result in all_results:
        total_metrics = result[0]
        percentile_metrics = result[1]
        percentiles = percentile_metrics[PercentileMetrics.PERCENTILES]
        try:
            concurrency = total_metrics.get(Metrics.NUMBER_OF_CONCURRENCY, 0)
            rps = total_metrics.get(Metrics.REQUEST_THROUGHPUT, 0)
            avg_latency = total_metrics.get(Metrics.AVERAGE_LATENCY, 0)
            p99_latency = percentile_metrics.get(PercentileMetrics.LATENCY)[percentiles.index('99%')]
            avg_tps = total_metrics.get(Metrics.OUTPUT_TOKEN_THROUGHPUT, 0)
            avg_ttft = total_metrics.get(Metrics.AVERAGE_TIME_TO_FIRST_TOKEN, 0)
            p99_ttft = percentile_metrics.get(PercentileMetrics.TTFT)[percentiles.index('99%')]
            success_rate = (
                total_metrics.get(Metrics.SUCCEED_REQUESTS, 0) / total_metrics.get(Metrics.TOTAL_REQUESTS, 1)
            ) * 100
            avg_tpot = total_metrics.get(Metrics.AVERAGE_TIME_PER_OUTPUT_TOKEN, 0)
            p99_tpot = percentile_metrics.get(PercentileMetrics.TPOT)[percentiles.index('99%')]

            # Ensure all values are valid numbers
            if any(x is None for x in [concurrency, rps, avg_latency, p99_latency, avg_tps, avg_ttft]):
                logger.warning(f'Warning: Test results for concurrency {concurrency} contain invalid data, skipped')
                continue

            summary.append([
                concurrency,
                f'{rps:.2f}' if rps is not None else 'N/A',
                f'{avg_latency:.3f}' if avg_latency is not None else 'N/A',
                f'{p99_latency:.3f}' if p99_latency is not None else 'N/A',
                f'{avg_tps:.2f}' if avg_tps is not None else 'N/A',
                f'{avg_ttft:.3f}' if avg_ttft is not None else 'N/A',
                f'{success_rate:.1f}%' if success_rate is not None else 'N/A',
                f'{p99_ttft:.3f}' if p99_ttft is not None else 'N/A',
                f'{avg_tpot:.3f}' if avg_tpot is not None else 'N/A',
                f'{p99_tpot:.3f}' if p99_tpot is not None else 'N/A',
            ])

            total_tokens += total_metrics.get(Metrics.AVERAGE_OUTPUT_TOKENS_PER_REQUEST,
                                              0) * total_metrics.get(Metrics.SUCCEED_REQUESTS, 0)
            total_time += total_metrics.get(Metrics.TIME_TAKEN_FOR_TESTS, 0)
        except Exception as e:
            logger.warning(
                f"Warning: Error processing results for concurrency {result.get('concurrency', 'unknown')}: {str(e)}"
            )
            continue

    if not summary:
        logger.warning('Error: No valid test result data')
        return [], 0, 0

    return summary, total_tokens, total_time


def print_summary(all_results, model_name):
    """Print test results summary"""
    summary, total_tokens, total_time = analyze_results(all_results)

    if not summary:
        logger.warning('No available test result data to display')
        return

    console = Console(width=100)  # Set fixed width

    # Create title panel
    title = Text('Performance Test Summary Report', style='bold')
    console.print(Panel(title, width=60))

    # Print basic information
    basic_info = Table(show_header=False, width=60)
    basic_info.add_column('Name', style='cyan', width=25)
    basic_info.add_column('Value', style='green', width=35)

    basic_info.add_row('Model', model_name)
    basic_info.add_row('Total Generated', f'{total_tokens:,} tokens')
    basic_info.add_row('Total Test Time', f'{total_time:.2f} seconds')
    basic_info.add_row('Avg Output Rate', f'{total_tokens / total_time:.2f} tokens/sec')

    console.print('\nBasic Information:')
    console.print(basic_info)

    # Create detailed performance metrics table
    table = Table(
        title='Detailed Performance Metrics',
        show_header=True,
        header_style='bold cyan',
        border_style='blue',
        width=100,  # Set total table width
        pad_edge=False,  # Reduce edge padding
        min_width=60,  # Minimum width
    )

    # Add columns (set fixed column widths)
    table.add_column('Conc.', justify='right', style='cyan')
    table.add_column('RPS', justify='right')
    table.add_column('Avg Lat.(s)', justify='right')
    table.add_column('P99 Lat.(s)', justify='right')
    table.add_column('Gen. toks/s', justify='right')
    table.add_column('Avg TTFT(s)', justify='right')
    table.add_column('P99 TTFT(s)', justify='right')
    table.add_column('Avg TPOT(s)', justify='right')
    table.add_column('P99 TPOT(s)', justify='right')
    table.add_column('Success Rate', justify='right', style='green')

    # Add data rows
    for row in summary:
        try:
            # Set row style based on success rate
            success_rate = float(row[6].rstrip('%'))
            row_style = 'green' if success_rate >= 95 else 'yellow' if success_rate >= 80 else 'red'

            table.add_row(
                str(row[0]),  # Concurrency
                f'{float(row[1]):.2f}',  # RPS
                f'{float(row[2]):.3f}',  # Average Latency
                f'{float(row[3]):.3f}',  # P99 Latency
                f'{float(row[4]):.2f}',  # Average TPS
                f'{float(row[5]):.3f}',  # First Token Latency
                f'{float(row[7]):.3f}',  # P99 TTFT
                f'{float(row[8]):.3f}',  # Average TPOT
                f'{float(row[9]):.3f}',  # P99 TPOT
                row[6],  # Success Rate
                style=row_style
            )
        except ValueError as e:
            console.print(f'Warning: Error processing row data: {str(e)}', style='bold red')
            continue

    console.print('\n')
    console.print(table)

    # Calculate and display best performance configuration
    try:
        best_rps_idx = np.argmax([float(row[1]) if row[1] != 'N/A' else -1 for row in summary])
        best_latency_idx = np.argmin([float(row[2]) if row[2] != 'N/A' else float('inf') for row in summary])

        perf_info = Table(title='Best Performance Configuration', show_header=False, box=None, width=60)
        perf_info.add_column('Metric', style='cyan', width=20)
        perf_info.add_column('Value', style='green', width=40)

        perf_info.add_row('Highest RPS', f'Concurrency {summary[best_rps_idx][0]} ({summary[best_rps_idx][1]} req/sec)')
        perf_info.add_row(
            'Lowest Latency', f'Concurrency {summary[best_latency_idx][0]} ({summary[best_latency_idx][2]} seconds)'
        )

        console.print('\n')
        console.print(perf_info)

        # Performance recommendations
        recommendations = []
        if best_rps_idx == len(summary) - 1:
            recommendations.append(
                'The system seems not to have reached its performance bottleneck, try higher concurrency'
            )
        elif best_rps_idx == 0:
            recommendations.append('Consider lowering concurrency, current load may be too high')
        else:
            recommendations.append(f'Optimal concurrency range is around {summary[best_rps_idx][0]}')

        success_rate = float(summary[-1][6][:-1])
        if success_rate < 95:
            recommendations.append(
                'Success rate is low at high concurrency, check system resources or reduce concurrency'
            )

        recommend_text = Text('\nPerformance Recommendations:', style='bold cyan')
        console.print(recommend_text)
        for rec in recommendations:
            console.print(f'• {rec}', style='yellow')

    except Exception as e:
        console.print(f'Warning: Error generating performance analysis: {str(e)}', style='bold red')



import base64
import os

from evalscope.api.dataset import Sample
from evalscope.api.evaluator import TaskState
from evalscope.api.messages.chat_message import ChatMessageUser
from evalscope.api.messages.content import ContentImage
from evalscope.api.metric import Score
from evalscope.api.model import ChatCompletionChoice, Model, ModelOutput
from evalscope.api.registry import get_metric
from evalscope.constants import EvalType, FileConstants
from evalscope.utils import get_logger
from evalscope.utils.function_utils import thread_safe
from .default_data_adapter import DefaultDataAdapter

logger = get_logger()


class Text2ImageAdapter(DefaultDataAdapter):
    """Text to Image Adapter for benchmarks."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.add_aggregation_name = False  # Do not add aggregation name in the report by default

    def load_from_disk(self, **kwargs):
        return super().load_from_disk(use_local_loader=True)

    def record_to_sample(self, record) -> Sample:
        """Convert a record dictionary to a Sample object."""
        return Sample(
            input=[ChatMessageUser(content=record['prompt'])],
            metadata={
                'prompt': record['prompt'],
                'category': record.get('category', ''),
                'tags': record.get('tags', []),
                FileConstants.ID: record[FileConstants.ID],
                FileConstants.IMAGE_PATH: record.get(FileConstants.IMAGE_PATH,
                                                     ''),  # Optional field for existing image path
            }
        )

    def _on_inference(self, model: Model, sample: Sample) -> ModelOutput:
        """
        Hook method called during the actual inference process.

        This method executes the model inference and can be overridden
        to implement custom inference logic or model interaction patterns.

        Args:
            model (Model): The model to use for inference
            sample (Sample): The sample to process

        Returns:
            ModelOutput: The raw output from the model
        """
        if self.eval_type == EvalType.MOCK_LLM:
            return ModelOutput(
                model=model.name,
                choices=[ChatCompletionChoice.from_content('')],
            )
        else:
            # Execute model inference with the processed input and any tools
            model_output = model.generate(input=sample.input, tools=sample.tools)
            return model_output

    def _on_inference_end(
        self, model: Model, sample: Sample, model_output: ModelOutput, output_dir: str, **kwargs
    ) -> TaskState:
        """
        Hook method called after inference completes. Save generated images to output_dir.

        Args:
            model (Model): The model that performed inference
            sample (Sample): The processed sample
            model_output (ModelOutput): The raw model output
            output_dir (str): The directory where the model output was saved

        Returns:
            TaskState: Complete state object for the inference task
        """
        if self.eval_type == EvalType.MOCK_LLM:
            return TaskState(
                model=model.name,
                sample=sample,
                messages=[model_output.message],
                output=model_output,
                completed=True,
            )
        else:
            image_id = f'{sample.metadata.get(FileConstants.ID, sample.id)}_{sample.group_id}'
            output_path = os.path.join(output_dir, 'images', f'{image_id}.png')
            if not os.path.exists(os.path.dirname(output_path)):
                os.makedirs(os.path.dirname(output_path))
            # get base64 image from model_output
            content = model_output.message.content[0]

            assert isinstance(content, ContentImage), 'Expected ContentImage in model output'

            image_base64 = content.image
            with open(output_path, 'wb') as f:
                f.write(base64.b64decode(image_base64))

            sample.metadata[FileConstants.IMAGE_PATH] = output_path
            return TaskState(
                model=model.name,
                sample=sample,
                messages=[model_output.message],
                output=model_output,
                completed=True,
            )

    # NOTE: thread safe is needed, since we can't batch inference here.
    @thread_safe
    def match_score(
        self, original_prediction: str, filtered_prediction: str, reference: str, task_state: TaskState
    ) -> Score:
        # Get prediction and prompt from task state
        image_path = task_state.metadata.get(FileConstants.IMAGE_PATH, original_prediction)
        prompt = task_state.input[0].content
        meta = task_state.metadata

        # Initialize the score object with prediction details
        score = Score(
            extracted_prediction=image_path,
            prediction=image_path,
        )

        # Calculate scores for each configured metric
        for metric in self.metric_list:
            try:
                if isinstance(metric, str):
                    metric_name = metric
                    metric_scorer = get_metric(metric)  # Get metric implementation from registry
                    metric_func = metric_scorer()  # Instantiate the metric scorer
                elif isinstance(metric, dict):
                    metric_name = list(metric.keys())[0]
                    metric_cls = get_metric(metric_name)
                    metric_func = metric_cls(**metric[metric_name])  # Initialize with parameters
                metric_score = metric_func(image_path, prompt)[0]

                # fine-granular metrics
                category = meta.get('category')
                if category:
                    metric_name = f'{metric_name}_{category}'
                if isinstance(metric_score, dict):
                    for k, v in metric_score.items():
                        score.value[f'{metric_name}_{k}'] = v.cpu().item()
                else:
                    score.value[metric_name] = metric_score.cpu().item()
            except Exception as e:
                logger.error(f'Error calculating metric {metric}: {e}')
                score.value[metric_name] = 0
                score.metadata[metric_name] = f'error: {str(e)}'

        return score



# Copyright (c) Alibaba, Inc. and its affiliates.
from typing import TYPE_CHECKING

from evalscope.utils.import_utils import _LazyModule

if TYPE_CHECKING:
    from .model_apis import llm_ckpt, mockllm, openai_api

else:
    _import_structure = {
        'model_apis': [
            'openai_api',
            'mockllm',
            'llm_ckpt',
        ]
    }

    import sys

    sys.modules[__name__] = _LazyModule(
        __name__,
        globals()['__file__'],
        _import_structure,
        module_spec=__spec__,
        extra_objects={},
    )



from abc import ABC, abstractmethod
from typing import Callable, Iterable, List, Union

from evalscope.utils import get_logger
from evalscope.utils.function_utils import thread_safe

logger = get_logger()


class Metric(ABC):
    """
    Metric classes operate on a sample level.
    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Can define custom behavior here, if an individual instantiation of a Metric class should have state.
        """

    @abstractmethod
    def apply(self, predictions: List[str], references: List[str]) -> List[float]:
        pass

    def __call__(self, prediction: str, reference: str) -> float:
        """
        Allows the metric to be called like a function.
        """
        return self.apply([prediction], [reference])[0]


class T2IMetric(Metric):
    _instance = None

    @thread_safe
    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, *args, **kwargs):
        cls = self.__class__
        if hasattr(self, '_init_done'):
            return
        logger.info(f'Initializing {cls.__name__}...')
        self._init_once(*args, **kwargs)
        self._init_done = True

    def _init_once(self, *args, **kwargs):
        pass

    def apply(self, images: List[str], texts: List[str], **kwargs) -> List[Union[float, dict]]:
        pass

    def __call__(self, image: str, text: str, **kwargs) -> Union[float, dict]:
        return self.apply([image], [text], **kwargs)[0]



import ast
import json
import os
import random
import regex as re
from typing import Union

from evalscope.utils.logger import get_logger

logger = get_logger()


def fix_json(input_str):
    # Add double quotes around keys using regex
    fixed_str = re.sub(r'(\w+):', r'"\1":', input_str)

    # Add double quotes around string values if necessary and wrap int/float values in []
    def format_value(match):
        key, value, comma = match.groups()
        value = value.strip()
        # Check if value is an integer or float
        if re.match(r'^-?\d+(\.\d+)?$', value):
            value = f'[{value}]'
        # Check if value is a boolean or null
        elif re.match(r'^(true|false|null)$', value, re.IGNORECASE):
            pass  # leave as is
        else:
            # Add quotes around string values
            value = f'"{value}"'
        return f'{key}: {value}{comma}'

    fixed_str = re.sub(r'(".*?"):(.*?)(,|})', format_value, fixed_str)

    return fixed_str


def read_file_to_string(file_path):
    """
    Reads the contents of a text file and returns it as a string.

    :param file_path: The path to the text file.
    :return: A string containing the contents of the file.
    """
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            return file.read()
    except FileNotFoundError:
        logger.info(f'The file {file_path} was not found.')
        return None
    except Exception as e:
        logger.info(f'An error occurred: {e}')
        return None


def read_files_to_string(file_paths):
    """
    Reads the contents of multiple text files and returns them as a single string,
    with each file's contents separated by a newline.

    :param file_paths: A list of paths to text files.
    :return: A string containing the concatenated contents of the files.
    """
    all_contents = []  # List to hold the contents of each file

    for file_path in file_paths:
        try:
            with open(file_path, 'r', encoding='utf-8') as file:
                all_contents.append(file.read())
        except FileNotFoundError:
            logger.info(f'The file {file_path} was not found.')
        except Exception as e:
            logger.info(f'An error occurred while reading {file_path}: {e}')

    # Join all the contents with a newline character
    return '\n'.join(all_contents)


def get_file_path(filename: Union[str, os.PathLike], search_from: Union[str, os.PathLike] = '.'):
    """
    Search for a file across a directory and return its absolute path.

    Args:
        filename (Union[str, os.PathLike]): The name of the file to search for.
        search_from (Union[str, os.PathLike], optional): The directory from which to start the search. Defaults to ".".

    Returns:
        str: Absolute path to the found file.

    Raises:
        FileNotFoundError: If the file is not found.
    """
    for root, dirs, files in os.walk(search_from):
        for name in files:
            if name == filename:
                return os.path.abspath(os.path.join(root, name))
    raise FileNotFoundError(filename, 'not found.')


# +=========================================================================================
def verify(s, target_sequence):
    # Count the occurrences of the target sequence
    count = s.count(target_sequence)

    # Check if the target sequence appears exactly twice
    return count == 2


def is_int_between_0_and_10(s):
    try:
        num = int(s)
        return 0 <= num <= 10
    except ValueError:
        return False


def is_str_a_list_of_ints_0_to_10(s):
    try:
        # Attempt to parse the string as a Python literal (list, dict, etc.)
        parsed = ast.literal_eval(s)

        # Check if the parsed object is a list
        if not isinstance(parsed, list):
            return False

        # Check if all elements are integers and between 0 to 10
        return all(isinstance(item, int) and 0 <= item <= 10 for item in parsed)

    except (ValueError, SyntaxError):
        # If parsing fails or any other error occurs
        return False


def is_str_valid_score_format_brackets(s):
    try:
        # Removing brackets and splitting the string by commas
        content = s.strip('[]').split(',')

        length = len(content)

        # Parsing each element and checking the format and range
        scores = {}
        for item in content:
            key, value = item.split(':')
            key = key.strip()
            value = int(value.strip())

            # Check if the key starts with 'score' and the value is in the correct range
            if not key.startswith('score') or not 0 <= value <= 10:
                return False

            scores[key] = value

        fetch_words = [f'score{i+1}' for i in range(length)]
        # Check if at least 'score1' and 'score2' are present
        return all(key in scores for key in fetch_words)

    except (ValueError, SyntaxError):
        # If any parsing error occurs
        return False


# +=========================================================================================
def mllm_output_to_dict(input_string, give_up_parsing=False):
    """
    Args:
        input_string (str): actually the output of the mllm model to be parsed
        output_file_name (str): The name of the output file.
    """
    # Catch for gpt4v rate_limit_exceeded error
    if input_string == 'rate_limit_exceeded':
        return 'rate_limit_exceeded'

    # Define the delimiters
    delimiter = '||V^=^V||'

    if input_string.count(delimiter) == 2:
        if not verify(input_string, delimiter):
            logger.info('The required delimiters were not found correctly in the string.')
            return False
        # Extract the content between the delimiters
        start_index = input_string.find(delimiter) + len(delimiter)
        end_index = input_string.rfind(delimiter)
    else:
        # find the json mannually
        # some mllm tends not to output the delimiters, but it does output the json contents
        # so we will find the json content mannually
        start_index = input_string.find('{')
        end_index = input_string.rfind('}') + 1
        if start_index == -1 or end_index == 0:
            # json not found
            # some mllm tends to output only a list of scores like [6, 0],
            # this time we will just get the scores and ignore the reasoning (other part of the json)
            start_index = input_string.find('[')
            end_index = input_string.rfind(']') + 1
            if give_up_parsing:  # if we want to give up parsing
                guessed_value = random.randint(0, 10)
                logger.info(f'Failed to find the json content in the string. Guess a value : {guessed_value}.')
                json_content = {'score': [guessed_value], 'reasoning': f'guess_if_cannot_parse | {input_string}'}
                json_str = json.dumps(json_content)
                input_string = json_str
                start_index = 0
                end_index = len(json_str)
            elif re.match(r'^\[\d+, ?\d+\]$', input_string[start_index:end_index]):
                scores = json.loads(input_string[start_index:end_index])
                if not isinstance(scores, list):
                    scores = [scores]
                json_content = {'score': scores, 'reasoning': 'System: output is simply a list of scores'}
                json_str = json.dumps(json_content)
                input_string = json_str
                start_index = 0
                end_index = len(json_str)
            elif is_int_between_0_and_10(input_string):  # if output is simply a number
                scores = [int(input_string)]
                json_content = {'score': scores, 'reasoning': 'System: output is simply a number'}
                json_str = json.dumps(json_content)
                input_string = json_str
                start_index = 0
                end_index = len(json_str)
            else:
                logger.info('Failed to find the json content in the string.')
                return False

    # Check if we found two delimiters
    if start_index != -1 and end_index != -1 and start_index != end_index:
        # Extract the JSON string
        json_str = input_string[start_index:end_index].strip()
        json_str = json_str.replace('\n', '')
        # Parse the JSON string into a dictionary
        try:
            new_data = json.loads(json_str)
            if not isinstance(new_data['score'], list):
                new_data['score'] = [new_data['score']]
        except Exception:
            logger.info('Now fixing: ', json_str)
            try:
                new_data = json.loads(fix_json(json_str))
                return new_data
            except Exception:
                logger.info('Error: Cannot fix', json_str)
                return False
        return new_data
    else:
        logger.info('The required delimiters were not found correctly in the string.')
        return False


def write_entry_to_json_file(input_string, uid, prompt_input, vision_input, output_file_name, give_up_parsing=False):
    """
    Args:
        input_string (str): actually the output of the mllm model to be parsed
        uid (str): The unique identifier for the each item in the test data
        prompt_input (str): The prompt input for the entry. text prompt.
        vision_input (str): The vision input for the entry. image links.
        output_file_name (str): The name of the output file.
    """
    # Catch for gpt4v rate_limit_exceeded error
    if input_string == 'rate_limit_exceeded':
        return 'rate_limit_exceeded'

    # Define the delimiters
    delimiter = '||V^=^V||'

    if input_string.count(delimiter) == 2:
        if not verify(input_string, delimiter):
            logger.info('The required delimiters were not found correctly in the string.')
            return False
        # Extract the content between the delimiters
        start_index = input_string.find(delimiter) + len(delimiter)
        end_index = input_string.rfind(delimiter)
    else:
        # find the json mannually
        # some mllm tends not to output the delimiters, but it does output the json contents
        # so we will find the json content mannually
        start_index = input_string.find('{')
        end_index = input_string.rfind('}') + 1
        if start_index == -1 or end_index == 0:
            # json not found
            # some mllm tends to output only a list of scores like [6, 0],
            # this time we will just get the scores and ignore the reasoning (other part of the json)
            start_index = input_string.find('[')
            end_index = input_string.rfind(']') + 1
            if give_up_parsing:  # if we want to give up parsing
                guessed_value = random.randint(0, 10)
                logger.info(f'Failed to find the json content in the string. Guess a value : {guessed_value}.')
                json_content = {'score': [guessed_value], 'reasoning': f'guess_if_cannot_parse | {input_string}'}
                json_str = json.dumps(json_content)
                input_string = json_str
                start_index = 0
                end_index = len(json_str)
            elif re.match(r'^\[\d+, ?\d+\]$', input_string[start_index:end_index]):
                scores = json.loads(input_string[start_index:end_index])
                json_content = {'score': scores, 'reasoning': None}
                json_str = json.dumps(json_content)
                input_string = json_str
                start_index = 0
                end_index = len(json_str)
            elif is_int_between_0_and_10(input_string):  # if output is simply a number
                scores = [int(input_string)]
                json_content = {'score': scores, 'reasoning': None}
                json_str = json.dumps(json_content)
                input_string = json_str
                start_index = 0
                end_index = len(json_str)
            else:
                logger.info('Failed to find the json content in the string.')
                return False

    # Check if we found two delimiters
    if start_index != -1 and end_index != -1 and start_index != end_index:
        # Extract the JSON string
        json_str = input_string[start_index:end_index].strip()
        json_str = json_str.replace('\n', '')
        try:
            # Parse the JSON string into a dictionary
            new_data = json.loads(json_str)

            # Ensure the directory exists
            os.makedirs(os.path.dirname(output_file_name), exist_ok=True)

            # Initialize or load existing data
            if os.path.exists(output_file_name):
                with open(output_file_name, 'r') as json_file:
                    data = json.load(json_file)
            else:
                data = {}

            # If the additional key is already in the data, add or update notes
            if uid in data:
                data[uid].update(new_data)  # Update with new data
                if prompt_input:  # If there are new notes, update or add them
                    data[uid]['prompt_input'] = prompt_input
                if vision_input:  # If there are new notes, update or add them
                    data[uid]['vision_input'] = vision_input
            else:
                # If it's a new key, add the entry to the dictionary
                data[uid] = new_data
                if prompt_input:
                    data[uid]['prompt_input'] = prompt_input
                if vision_input:
                    data[uid]['vision_input'] = vision_input

            # Write the updated data to the file
            with open(output_file_name, 'w') as json_file:
                json.dump(data, json_file, indent=4)

            logger.info(f'Data was successfully updated in {output_file_name}')
            return True
        except json.JSONDecodeError as e:
            logger.info(f'An error occurred while parsing the JSON content: {e}')
            return False
    else:
        logger.info('The required delimiters were not found correctly in the string.')
        return False


def check_key_in_json(file_path, key):
    try:
        with open(file_path, 'r') as json_file:
            data = json.load(json_file)

        # Check if the key exists at the top level of the JSON structure
        if key in data:
            return True
        else:
            return False
    except FileNotFoundError:
        logger.info(f'The file {file_path} was not found.')
    except json.JSONDecodeError as e:
        logger.info(f'Error reading {file_path}: {e}')
    except Exception as e:
        logger.info(f'An error occurred with {file_path}: {e}')
    return False



import ast
import re
from typing import Any, Dict, List

from evalscope.api.benchmark import BenchmarkMeta, VisionLanguageAdapter
from evalscope.api.dataset import Sample
from evalscope.api.evaluator import TaskState
from evalscope.api.messages import ChatMessageUser, Content, ContentImage, ContentText
from evalscope.api.registry import register_benchmark
from evalscope.constants import Tags
from evalscope.utils.io_utils import bytes_to_base64
from evalscope.utils.logger import get_logger
from evalscope.utils.multi_choices import MultipleChoiceTemplate, parse_answers, prompt

# flake8: noqa

logger = get_logger()

SUBSET_LIST = [
    'Accounting',
    'Agriculture',
    'Architecture_and_Engineering',
    'Art',
    'Art_Theory',
    'Basic_Medical_Science',
    'Biology',
    'Chemistry',
    'Clinical_Medicine',
    'Computer_Science',
    'Design',
    'Diagnostics_and_Laboratory_Medicine',
    'Economics',
    'Electronics',
    'Energy_and_Power',
    'Finance',
    'Geography',
    'History',
    'Literature',
    'Manage',
    'Marketing',
    'Materials',
    'Math',
    'Mechanical_Engineering',
    'Music',
    'Pharmacy',
    'Physics',
    'Psychology',
    'Public_Health',
    'Sociology',
]

MULT_CHOICE_PROMPT = MultipleChoiceTemplate.SINGLE_ANSWER_COT

OPEN_PROMPT = """
Solve the following problem step by step. The last line of your response should be of the form "ANSWER: $ANSWER" (without quotes) where $ANSWER is the answer to the problem.

{question}

Remember to put your answer on its own line at the end in the form "ANSWER: $ANSWER" (without quotes) where $ANSWER is the answer to the problem, and you do not need to use a \\boxed command.
"""

MULTI_CHOICE_TYPE = 'multiple-choice'
OPEN_TYPE = 'open'


@register_benchmark(
    BenchmarkMeta(
        name='mmmu',
        pretty_name='MMMU',
        tags=[Tags.MULTI_MODAL, Tags.KNOWLEDGE, Tags.QA],
        description=
        'MMMU (A Massive Multi-discipline Multimodal Understanding and Reasoning Benchmark for Expert AGI) benchmark designed to evaluate multimodal models on massive multi-discipline tasks demanding college-level subject knowledge and deliberate reasoning. MMMU includes 11.5K meticulously collected multimodal questions from college exams, quizzes, and textbooks, covering six core disciplines: Art & Design, Business, Science, Health & Medicine, Humanities & Social Science, and Tech & Engineering. These questions span 30 subjects and 183 subfields, comprising 30 highly heterogeneous image types, such as charts, diagrams, maps, tables, music sheets, and chemical structures.',  # noqa: E501
        dataset_id='AI-ModelScope/MMMU',
        subset_list=SUBSET_LIST,
        metric_list=['acc'],
        eval_split='validation',
        prompt_template=OPEN_PROMPT,
    )
)
class MMMUAdapter(VisionLanguageAdapter):
    MAX_IMAGES: int = 7

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def record_to_sample(self, record: Dict[str, Any]) -> Sample:
        question_type = record['question_type']
        content_list, answers_list = MMMUAdapter.create_content_and_answers_list(record)

        metadata = {
            'id': record['id'],
            'question_type': record['question_type'],
            'subfield': record['subfield'],
            'explanation': record['explanation'],
            'img_type': record['img_type'],
            'topic_difficulty': record['topic_difficulty'],
        }

        if question_type == MULTI_CHOICE_TYPE:
            return Sample(
                input=[ChatMessageUser(content=content_list)],
                choices=answers_list,
                target=record['answer'],
                metadata=metadata,
            )
        elif question_type == OPEN_TYPE:
            return Sample(
                input=[ChatMessageUser(content=content_list)],
                target=record['answer'],
                metadata=metadata,
            )
        else:
            raise ValueError(f'Unsupported question type: {question_type}')

    def extract_answer(self, prediction: str, task_state: TaskState) -> str:
        question_type = task_state.metadata['question_type']
        if question_type == MULTI_CHOICE_TYPE:
            answers = parse_answers(task_state)
            return ''.join(sorted(list(answers)))
        elif question_type == OPEN_TYPE:
            pattern = r'ANSWER:\s*(.*)'
            match = re.search(pattern, prediction)
            if match:
                return match.group(1).strip()
            return prediction.strip()
        else:
            raise ValueError(f'Unsupported question type: {question_type}')

    @staticmethod
    def create_content_and_answers_list(record: Dict[str, Any]) -> tuple[List[Content], List[str]]:
        """
        Create a list of content elements and a list of answers from a record.

        Args:
            record (dict): The record containing question, images, and options.


        Returns:
            tuple: A tuple containing:
                - content_list (list): A list of content elements (text and images).
                - answers_list (list): A list of possible answers (for multiple-choice questions).
        """
        question_type = record['question_type']

        if question_type == MULTI_CHOICE_TYPE:
            answers_list: List[str] = ast.literal_eval(record['options'])
            input_text = prompt(question=record['question'], choices=answers_list, template=MULT_CHOICE_PROMPT)
            content_list: List[Content] = [ContentText(text=input_text)]
        else:
            answers_list: List[str] = []
            content_list: List[Content] = [ContentText(text=OPEN_PROMPT.format(question=record['question']))]

        for i in range(MMMUAdapter.MAX_IMAGES):
            image = record[f'image_{i+1}']
            if image:
                image_base64 = bytes_to_base64(image['bytes'], format='png', add_header=True)
                content_list.append(ContentImage(image=image_base64))

        return content_list, answers_list



# Copyright (c) Alibaba, Inc. and its affiliates.
import os

from evalscope.api.benchmark import BenchmarkMeta, Text2ImageAdapter
from evalscope.api.dataset import Sample
from evalscope.api.messages import ChatMessageUser
from evalscope.api.registry import get_metric, register_benchmark
from evalscope.constants import Tags
from evalscope.utils.logger import get_logger

logger = get_logger()


@register_benchmark(
    BenchmarkMeta(
        name='genai_bench',
        pretty_name='GenAI-Bench',
        dataset_id='AI-ModelScope/T2V-Eval-Prompts',
        description='GenAI-Bench Text-to-Image Benchmark. Includes 1600 prompts for text-to-image task.',
        tags=[Tags.TEXT_TO_IMAGE],
        subset_list=['GenAI-Bench-1600'],
        metric_list=['VQAScore'],
        few_shot_num=0,
        train_split=None,
        eval_split='test',
    )
)
class GenAIBenchAdapter(Text2ImageAdapter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def load_from_disk(self, **kwargs):
        if os.path.isfile(self.dataset_id):
            file_name = os.path.basename(self.dataset_id)
            file_without_ext = os.path.splitext(file_name)[0]
            self.subset_list = [file_without_ext]

        return super().load_from_disk(use_local_loader=True)

    def record_to_sample(self, record) -> Sample:
        """Convert a record dictionary to a Sample object."""
        advanced = record['tags'].get('advanced')
        return Sample(
            input=[ChatMessageUser(content=record['prompt'])],
            metadata={
                'id': record['id'],
                'prompt': record['prompt'],
                'category': 'advanced' if advanced else 'basic',
                'tags': record.get('tags', []),
                'image_path': record.get('image_path', ''),  # Optional field for existing image path
            }
        )
